# -*- coding: utf-8 -*-
"""
CommandLine;
    # Reset IBEIS database (can skip if done)
    python -m wbia.tests.reset_testdbs --reset_mtest
    python -m wbia --tf reset_mtest

Notes:
    Moving components: java, tomcat, wildbook.war.

    python -m utool.util_inspect check_module_usage --pat="manual_wildbook_funcs.py"


CommandLine;
    # Start IA server
    python -m wbia --web --db PZ_MTEST

    # Reset Wildbook database
    python -m wbia purge_local_wildbook

    # Install Wildbook
    python -m wbia install_wildbook

    # Startup Wildbook
    python -m wbia startup_wildbook_server
    --show

    # Poll wildbook info
    python -m wbia get_wildbook_ia_url

    # Login to wildbook (can skip)
    python -m wbia test_wildbook_login

    # Ship ImageSets to wildbook
    python -m wbia wildbook_signal_imgsetid_list

    # Change annotations names to a single name
    python -m wbia wildbook_signal_annot_name_changes:1

    # Change annotations names back to normal
    python -m wbia wildbook_signal_annot_name_changes:2
"""
import logging
import utool as ut
import requests
from wbia.control import controller_inject
from wbia.control import wildbook_manager as wb_man  # NOQA
from wbia.control.controller_inject import make_ibs_register_decorator

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


DISABLE_WILDBOOK_SIGNAL = ut.get_argflag('--no-wb-signal')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


register_api = controller_inject.get_wbia_flask_api(__name__)


# PREFERED_BROWSER = 'chrome'
# webbrowser._tryorder
PREFERED_BROWSER = None
if ut.get_computer_name() == 'hyrule':
    PREFERED_BROWSER = 'firefox'


@register_ibs_method
def get_wildbook_base_url(ibs, wb_target=None):
    if DISABLE_WILDBOOK_SIGNAL:
        message = 'Wildbook signals are turned off via the command line'
        logger.info(message)
        raise IOError(message)

    wb_port = 8080
    wb_target = ibs.const.WILDBOOK_TARGET if wb_target is None else wb_target

    if ibs.containerized:
        wb_hostname = 'nginx'
        wb_target = ''
    else:
        wb_hostname = '127.0.0.1'

    wb_hostname = str(wb_hostname)
    wb_port = str(wb_port)
    wb_target = str(wb_target)

    wildbook_base_url = 'http://%s:%s/%s/' % (wb_hostname, wb_port, wb_target)
    wildbook_base_url = wildbook_base_url.strip('/')

    logger.info('USING WB BASEURL: %r' % (wildbook_base_url,))
    return wildbook_base_url


@register_ibs_method
def assert_ia_available_for_wb(ibs, wb_target=None):
    # Test if we have a server alive
    try:
        ia_url = ibs.get_wildbook_ia_url(wb_target)

        if ia_url is None:
            message = 'Wildbook signals are turned off via the command line'
            logger.info(message)
            raise IOError(message)
    except IOError:
        logger.info('[ibs.assert_ia_available_for_wb] Caught IOError, returning None')
        return None
    except Exception as ex:
        ut.printex(ex, 'Could not get IA url. BLINDLY CHARCHING FORWARD!', iswarning=True)
    else:
        have_server = False
        for count in ut.delayed_retry_gen([1], timeout=3, raise_=False):
            try:
                rsp = requests.get(ia_url + '/api/test/heartbeat/', timeout=3)
                have_server = rsp.status_code == 200
                break
            except requests.ConnectionError:
                pass
        if not have_server:
            raise Exception('The image analysis server is not started.')
        return have_server


@register_ibs_method
def get_wildbook_ia_url(ibs, wb_target=None):
    """
    Where does wildbook expect us to be?

    CommandLine:
        python -m wbia get_wildbook_ia_url

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_wildbook_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> ia_url = ibs.get_wildbook_ia_url()
        >>> print('ia_url = %r' % (ia_url,))
    """
    import requests

    try:
        wb_url = ibs.get_wildbook_base_url(wb_target)
    except IOError:
        logger.info('[ibs.get_wildbook_ia_url] Caught IOError, returning None')
        return None

    response = requests.get(wb_url + '/ia?status')
    status = response.status_code == 200
    if not status:
        raise Exception('Could not get IA status from wildbook')
    json_response = response.json()
    ia_url = json_response.get('iaURL')
    # logger.info('response = %r' % (response,))
    return ia_url


@register_ibs_method
@register_api('/api/wildbook/signal/annot/name/', methods=['PUT'])
def wildbook_signal_annot_name_changes(ibs, aid_list=None, wb_target=None, dryrun=False):
    r"""
    Args:
        aid_list (int):  list of annotation ids(default = None)
        tomcat_dpath (None): (default = None)
        wb_target (None): (default = None)
        dryrun (bool): (default = False)

    CommandLine:
        python -m wbia wildbook_signal_annot_name_changes:0 --dryrun
        python -m wbia wildbook_signal_annot_name_changes:1 --dryrun
        python -m wbia wildbook_signal_annot_name_changes:1
        python -m wbia wildbook_signal_annot_name_changes:2

    Setup:
        >>> wb_target = None
        >>> dryrun = ut.get_argflag('--dryrun')

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_wildbook_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> #gid_list = ibs.get_valid_gids()[0:10]
        >>> gid_list = ibs.get_valid_gids()[3:5]
        >>> aid_list = ut.flatten(ibs.get_image_aids(gid_list))
        >>> # Test case where some names change, some do not. There are no new names.
        >>> old_nid_list = ibs.get_annot_name_rowids(aid_list)
        >>> new_nid_list = ut.list_roll(old_nid_list, 1)
        >>> ibs.set_annot_name_rowids(aid_list, new_nid_list)
        >>> result = ibs.wildbook_signal_annot_name_changes(aid_list, wb_target, dryrun)
        >>> ibs.set_annot_name_rowids(aid_list, old_nid_list)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_wildbook_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> #gid_list = ibs.get_valid_gids()[0:10]
        >>> gid_list = ibs.get_valid_gids()[3:5]
        >>> aid_list = ut.flatten(ibs.get_image_aids(gid_list))
        >>> # Test case where all names change to one known name
        >>> #old_nid_list = ibs.get_annot_name_rowids(aid_list)
        >>> #new_nid_list = [old_nid_list[0]] * len(old_nid_list)
        >>> old_nid_list = [1, 2]
        >>> new_nid_list = [1, 1]
        >>> print('old_nid_list = %r' % (old_nid_list,))
        >>> print('new_nid_list = %r' % (new_nid_list,))
        >>> ibs.set_annot_name_rowids(aid_list, new_nid_list)
        >>> result = ibs.wildbook_signal_annot_name_changes(aid_list, wb_target, dryrun)
        >>> # Undo changes here (not undone in wildbook)
        >>> #ibs.set_annot_name_rowids(aid_list, old_nid_list)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_wildbook_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()[3:5]
        >>> aid_list = ut.flatten(ibs.get_image_aids(gid_list))
        >>> old_nid_list = [1, 2]
        >>> ibs.set_annot_name_rowids(aid_list, old_nid_list)
        >>> # Signal what currently exists (should put them back to normal)
        >>> result = ibs.wildbook_signal_annot_name_changes(aid_list, wb_target, dryrun)
    """
    logger.info(
        '[ibs.wildbook_signal_annot_name_changes] signaling annot name changes to wildbook'
    )
    try:
        wb_url = ibs.get_wildbook_base_url(wb_target)
    except IOError:
        logger.info(
            '[ibs.wildbook_signal_annot_name_changes] Caught IOError, returning None'
        )
        return None

    try:
        ibs.assert_ia_available_for_wb(wb_target)
    except Exception:
        pass
    if aid_list is None:
        aid_list = ibs.get_valid_aids(is_known=True)

    annot_uuid_list = ibs.get_annot_uuids(aid_list)
    annot_name_text_list = ibs.get_annot_name_texts(aid_list)
    grouped_uuids = ut.group_items(annot_uuid_list, annot_name_text_list)
    url = wb_url + '/ia'
    payloads = [
        {
            'resolver': {
                'assignNameToAnnotations': {
                    'name': new_name,
                    'annotationIds': ut.lmap(str, annot_uuids),
                }
            }
        }
        for new_name, annot_uuids in grouped_uuids.items()
    ]
    status_list = []
    for json_payload in ut.ProgressIter(payloads, lbl='submitting URL', freq=1):
        logger.info('[_send] URL=%r with json_payload=%r' % (url, json_payload))
        if dryrun:
            status = False
        else:
            response = requests.post(url, json=json_payload)
            status = response.status_code == 200
            if not status:
                logger.info('Failed to push new names')
                # logger.info(response.text)
        status_list.append(status)
    return status_list


@register_ibs_method
@register_api('/api/wildbook/signal/name/', methods=['PUT'])
def wildbook_signal_name_changes(
    ibs, nid_list, new_name_list, wb_target=None, dryrun=False
):
    r"""
    Args:
        nid_list (int):  list of name ids
        new_name_list (str):  list of corresponding names
        wb_target (None): (default = None)
        dryrun (bool): (default = False)

    CommandLine:
        python -m wbia wildbook_signal_name_changes:0 --dryrun
        python -m wbia wildbook_signal_name_changes:1 --dryrun
        python -m wbia wildbook_signal_name_changes:1
        python -m wbia wildbook_signal_name_changes:2

    Setup:
        >>> wb_target = None
        >>> dryrun = ut.get_argflag('--dryrun')
    """
    logger.info('[ibs.wildbook_signal_name_changes] signaling name changes to wildbook')
    try:
        wb_url = ibs.get_wildbook_base_url(wb_target)
    except IOError:
        logger.info('[ibs.wildbook_signal_name_changes] Caught IOError, returning None')
        return None

    try:
        ibs.assert_ia_available_for_wb(wb_target)
    except Exception:
        pass
    current_name_list = ibs.get_name_texts(nid_list)
    combined_list = sorted(list(zip(new_name_list, current_name_list)), reverse=True)
    url = wb_url + '/ia'
    json_payload = {
        'resolver': {
            'renameIndividuals': {
                'new': [_[0] for _ in combined_list],
                'old': [_[1] for _ in combined_list],
            }
        }
    }
    status_list = []
    logger.info('[_send] URL=%r with json_payload=%r' % (url, json_payload))
    if dryrun:
        status = False
    else:
        response = requests.post(url, json=json_payload)
        response_json = response.json()
        status = response.status_code == 200 and response_json['success']
        if not status:
            status_list = False
            logger.info('Failed to update names')
            # logger.info(response.text)
        else:
            for name_response in response_json['results']:
                status = name_response['success']
                error = name_response.get('error', '')
                status = status or 'unknown MarkedIndividual' in error
                status_list.append(status)
        # ut.embed()
    return status_list


@register_ibs_method
def wildbook_get_existing_names(ibs, wb_target=None):
    logger.info(
        '[ibs.wildbook_get_existing_names] getting existing names out of wildbook'
    )
    try:
        wb_url = ibs.get_wildbook_base_url(wb_target)
    except IOError:
        logger.info('[ibs.wildbook_get_existing_names] Caught IOError, returning None')
        return None

    try:
        ibs.assert_ia_available_for_wb(wb_target)
    except Exception:
        pass
    url = wb_url + '/rest/org.ecocean.MarkedIndividual'
    response = requests.get(url)
    response_json = response.json()
    try:
        wildbook_existing_name_list = [_['individualID'] for _ in response_json]
        wildbook_existing_name_list = list(set(wildbook_existing_name_list))
    except Exception:
        wildbook_existing_name_list = []
    return wildbook_existing_name_list


@register_ibs_method
@register_api('/api/wildbook/signal/imageset/', methods=['PUT'])
def wildbook_signal_imgsetid_list(
    ibs,
    imgsetid_list=None,
    set_shipped_flag=True,
    open_url_on_complete=True,
    wb_target=None,
    dryrun=False,
):
    """
    Exports specified imagesets to wildbook. This is a synchronous call.

    Args:
        imgsetid_list (list): (default = None)
        set_shipped_flag (bool): (default = True)
        open_url_on_complete (bool): (default = True)

    RESTful:
        Method: PUT
        URL:    /api/wildbook/signal/imageset/

    Ignore:
        cd $CODE_DIR/Wildbook/tmp

        # Ensure IA server is up
        python -m wbia --web --db PZ_MTEST

        # Reset IBEIS database
        python -m wbia.tests.reset_testdbs --reset_mtest
        python -m wbia  reset_mtest

        # Completely remove Wildbook database
        python -m wbia  purge_local_wildbook

        # Install Wildbook
        python -m wbia  install_wildbook

        # Startup Wildbook
        python -m wbia  startup_wildbook_server

        # Login to wildbook
        python -m wbia  test_wildbook_login

        # Ship ImageSets to wildbook
        python -m wbia  wildbook_signal_imgsetid_list

        # Change annotations names to a single name
        python -m wbia  wildbook_signal_annot_name_changes:1

        # Change annotations names back to normal
        python -m wbia  wildbook_signal_annot_name_changes:2

    CommandLine:
        python -m wbia wildbook_signal_imgsetid_list
        python -m wbia wildbook_signal_imgsetid_list --dryrun
        python -m wbia wildbook_signal_imgsetid_list --break

    SeeAlso:
        ~/local/build_scripts/init_wildbook.sh

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_wildbook_funcs import *  # NOQA
        >>> dryrun = ut.get_argflag('--dryrun')
        >>> wb_target = None
        >>> import wbia
        >>> # Need to start a web server for wildbook to hook into
        >>> defaultdb = 'PZ_MTEST'
        >>> ibs = wbia.opendb(defaultdb=defaultdb)
        >>> #gid_list = ibs.get_valid_gids()[0:10]
        >>> gid_list = ibs.get_valid_gids()[3:6]
        >>> new_imgsetid = ibs.create_new_imageset_from_images(gid_list)  # NOQA
        >>> imgsetid = new_imgsetid
        >>> print('new imageset uuid = %r' % (ibs.get_imageset_uuid(new_imgsetid),))
        >>> print('new imageset text = %r' % (ibs.get_imageset_text(new_imgsetid),))
        >>> imgsetid_list = [new_imgsetid]
        >>> ibs.set_imageset_processed_flags([new_imgsetid], [1])
        >>> gid_list = ibs.get_imageset_gids(new_imgsetid)
        >>> ibs.set_image_reviewed(gid_list, [1] * len(gid_list))
        >>> set_shipped_flag = True
        >>> open_url_on_complete = True
        >>> if ut.get_argflag('--bg'):
        >>>     with wbia.opendb_bg_web(defaultdb, managed=True) as web_ibs:
        ...         result = web_ibs.wildbook_signal_imgsetid_list(imgsetid_list, set_shipped_flag, open_url_on_complete, wb_target, dryrun)
        >>> else:
        ...     result = ibs.wildbook_signal_imgsetid_list(imgsetid_list, set_shipped_flag, open_url_on_complete, wb_target, dryrun)
        >>> # cleanup
        >>> #ibs.delete_imagesets(new_imgsetid)
        >>> print(result)

    """
    try:
        wb_url = ibs.get_wildbook_base_url(wb_target)
    except IOError:
        logger.info('[ibs.wildbook_signal_imgsetid_list] Caught IOError, returning None')
        return None

    try:
        ibs.assert_ia_available_for_wb(wb_target)
    except Exception:
        pass

    if imgsetid_list is None:
        imgsetid_list = ibs.get_valid_imgsetids()

    # Check to make sure imagesets are ok:
    for imgsetid in imgsetid_list:
        # First, check if imageset can be pushed
        aid_list = ibs.get_imageset_aids(imgsetid)
        assert (
            len(aid_list) > 0
        ), 'ImageSet imgsetid=%r cannot be shipped with0 annots' % (imgsetid,)
        unknown_flags = ibs.is_aid_unknown(aid_list)
        unnamed_aid_list = ut.compress(aid_list, unknown_flags)
        unnamed_ok_aid_list = ibs.filter_annots_general(unnamed_aid_list, minqual='ok')
        nUnnamedOk = sum(unnamed_ok_aid_list)
        assert nUnnamedOk == 0, (
            'ImageSet imgsetid=%r1 cannot be shipped becuase '
            'annotation(s) %r with an identifiable quality have '
            'not been named'
        ) % (imgsetid, unnamed_ok_aid_list)

    # Call Wildbook url to signal update
    logger.info(
        '[ibs.wildbook_signal_imgsetid_list] ship imgsetid_list = %r to wildbook'
        % (imgsetid_list,)
    )
    imageset_uuid_list = ibs.get_imageset_uuid(imgsetid_list)
    logger.info(
        '[ibs.wildbook_signal_imgsetid_list] ship imgset_uuid_list = %r to wildbook'
        % (imageset_uuid_list,)
    )

    url = wb_url + '/ia'
    dbname = ibs.db.get_db_init_uuid()
    occur_url_fmt = wb_url + '/occurrence.jsp?number={uuid}&dbname={dbname}'
    # enc_url_fmt = (wb_url + '/encounters/encounter.jsp?number={uuid}')

    # Check and push 'done' imagesets
    status_list = []
    for imgsetid, imageset_uuid in zip(imgsetid_list, imageset_uuid_list):
        logger.info('[_send] URL=%r' % (url,))
        json_payload = {'resolver': {'fromIAImageSet': str(imageset_uuid)}}
        if dryrun:
            status = False
        else:
            response = requests.post(url, json=json_payload)
            status = response.status_code == 200
            logger.info('response = %r' % (response,))
            if set_shipped_flag:
                ibs.set_imageset_shipped_flags([imgsetid], [status])
                if status and open_url_on_complete:
                    view_occur_url = occur_url_fmt.format(
                        uuid=imageset_uuid, dbname=dbname
                    )
                    _browser = ut.get_prefered_browser(PREFERED_BROWSER)
                    _browser.open_new_tab(view_occur_url)
        status_list.append(status)

    try:
        ibs.update_special_imagesets()
        ibs.notify_observers()
    except Exception:
        pass

    return status_list


@register_ibs_method
def get_flukebook_image_uuids(ibs):
    from datetime import datetime
    import uuid
    import pytz

    PST = pytz.timezone('US/Pacific')

    url = 'https://www.flukebook.org/acmIdSync.jsp'

    now = datetime.now(tz=PST)
    timestamp = now.strftime('%Y-%m-%d-%H-00-00')
    filename = 'flukebook.image.admid.%s.json' % (timestamp,)
    filepath = ut.grab_file_url(url, appname='wbia', fname=filename)

    with open(filepath, 'r') as file:
        file_content = file.read()
        file_json = ut.from_json(file_content)
    logger.info('Loaded %d Image ACM string UUIDs from Flukebook' % (len(file_json),))

    uuid_list = []
    for uuid_str in file_json:
        try:
            uuid_ = uuid.UUID(uuid_str)
            uuid_list.append(uuid_)
        except ValueError:
            continue

    logger.info('Validated %d Image UUIDs from Flukebook' % (len(uuid_list),))
    flukebook_image_uuid_list = list(set(uuid_list))
    logger.info(
        'Validated %d de-duplicated Image UUIDs from Flukebook' % (len(uuid_list),)
    )

    return flukebook_image_uuid_list


@register_ibs_method
def delete_flukebook_orphaned_image_uuids(ibs, auto_delete=True):
    flukebook_image_uuid_list = ibs.get_flukebook_image_uuids()

    gid_list = ibs.get_valid_gids()
    local_image_uuid_list = ibs.get_image_uuids(gid_list)

    unknown_uuid_list = list(set(flukebook_image_uuid_list) - set(local_image_uuid_list))
    candidate_uuid_list = list(
        set(local_image_uuid_list) - set(flukebook_image_uuid_list)
    )

    logger.info(
        'There are %d Image UUIDs in Flukebook that are not here'
        % (len(unknown_uuid_list),)
    )
    logger.info(
        'There are %d Image UUIDs in here that are not in Flukebook'
        % (len(candidate_uuid_list),)
    )

    if auto_delete and len(candidate_uuid_list) > 0:
        candidate_gid_list = ibs.get_image_gids_from_uuid(candidate_uuid_list)
        assert None not in candidate_gid_list
        ibs.delete_images(candidate_gid_list)

    return candidate_uuid_list


@register_ibs_method
def get_flukebook_annot_uuids(ibs, filter_match_against_on=True):
    from datetime import datetime
    import uuid
    import pytz

    PST = pytz.timezone('US/Pacific')

    url = 'https://www.flukebook.org/acmIdSync.jsp?annotations'

    now = datetime.now(tz=PST)
    timestamp = now.strftime('%Y-%m-%d-%H-00-00')
    filename = 'flukebook.annot.admid.%s.json' % (timestamp,)
    filepath = ut.grab_file_url(url, appname='wbia', fname=filename)

    with open(filepath, 'r') as file:
        file_content = file.read()
        file_json = ut.from_json(file_content)
    logger.info('Loaded %d Annot ACM string UUIDs from Flukebook' % (len(file_json),))

    uuid_list = []
    species_list = []
    for uuid_str in file_json:
        content = file_json[uuid_str]
        try:
            uuid_ = uuid.UUID(uuid_str)
        except ValueError:
            continue

        match = content.get('match', None)
        species = content.get('species', None)
        assert None not in [match, species]
        assert len(match) == len(species)

        if len(match) > 1:
            match_ = set(match)
            species_ = set(species)
            if len(match_) == 1 and len(species_) == 1:
                match = list(match_)
                species = list(species_)
                assert len(match) == len(species)

        if len(match) > 1:
            match_ = set(match)
            species_ = set(species)
            if len(species_) == 1:
                match = [any(match_)]
                species = list(species_)

        if len(match) == 1 and len(species) == 1:
            match = match[0]
            species = species[0]

            if filter_match_against_on is None or match == filter_match_against_on:
                uuid_list.append(uuid_)
                species_list.append(species)
        else:
            logger.info(match, species)

    assert len(uuid_list) == len(species_list)
    assert len(uuid_list) == len(set(uuid_list))
    logger.info('Validated %d Annotation UUIDs from Flukebook' % (len(uuid_list),))

    flukebook_annot_uuid_list = uuid_list
    flukebook_annot_species_list = species_list

    return flukebook_annot_uuid_list, flukebook_annot_species_list


@register_ibs_method
def delete_flukebook_orphaned_annot_uuids(ibs, auto_delete=True):
    from wbia import constants as const

    (
        flukebook_annot_uuid_list,
        flukebook_annot_species_list,
    ) = ibs.get_flukebook_annot_uuids()

    aid_list = ibs.get_valid_aids()
    local_annot_uuid_list = ibs.get_annot_uuids(aid_list)

    unknown_uuid_list = list(set(flukebook_annot_uuid_list) - set(local_annot_uuid_list))
    candidate_uuid_list = list(
        set(local_annot_uuid_list) - set(flukebook_annot_uuid_list)
    )

    logger.info(
        'There are %d Annot UUIDs in Flukebook that are not here'
        % (len(unknown_uuid_list),)
    )
    logger.info(
        'There are %d Annot UUIDs in here that are not in Flukebook'
        % (len(candidate_uuid_list),)
    )

    if auto_delete and len(candidate_uuid_list) > 0:
        candidate_aid_list = ibs.get_annot_aids_from_uuid(candidate_uuid_list)
        assert None not in candidate_aid_list
        ibs.delete_annots(candidate_aid_list)

    # Update species
    aid_list = ibs.get_valid_aids()
    local_annot_uuid_list = ibs.get_annot_uuids(aid_list)

    known_uuid_list = list(set(flukebook_annot_uuid_list) & set(local_annot_uuid_list))
    known_aid_list = ibs.get_annot_aids_from_uuid(known_uuid_list)
    assert None not in known_aid_list

    flukebook_species_dict = dict(
        zip(flukebook_annot_uuid_list, flukebook_annot_species_list)
    )
    known_species_list = ibs.get_annot_species(known_aid_list)

    flukebook_species_mapping = {
        'humpback_whale': 'megaptera_novaeangliae',
        'tursiops_sp.': 'tursiops_sp',
    }

    update_dict = {}
    update_aid_list = []
    update_species_list = []
    for known_aid, known_uuid, known_species in zip(
        known_aid_list, known_uuid_list, known_species_list
    ):
        flukebook_species = flukebook_species_dict.get(known_uuid, None)
        if flukebook_species is None:
            flukebook_species = const.UNKNOWN
        assert flukebook_species is not None
        flukebook_species = flukebook_species.lower()
        flukebook_species = flukebook_species.replace(' ', '_')
        flukebook_species = flukebook_species_mapping.get(
            flukebook_species, flukebook_species
        )

        if known_species != flukebook_species:

            if flukebook_species == const.UNKNOWN and known_species != const.UNKNOWN:
                continue

            update_aid_list.append(known_aid)
            update_species_list.append(flukebook_species)

            if known_species not in update_dict:
                update_dict[known_species] = {}
            if flukebook_species not in update_dict[known_species]:
                update_dict[known_species][flukebook_species] = 0
            update_dict[known_species][flukebook_species] += 1

    assert len(update_aid_list) == len(update_species_list)
    logger.info(ut.repr3(update_dict))

    ibs.set_annot_species(update_aid_list, update_species_list)

    return candidate_uuid_list, update_dict


@register_ibs_method
@register_api('/api/flukebook/sync/', methods=['GET', 'POST'])
def flukebook_sync(ibs, **kwargs):
    candidate_image_uuid_list = ibs.delete_flukebook_orphaned_image_uuids()
    candidate_annot_uuid_list, update_dict = ibs.delete_flukebook_orphaned_annot_uuids()

    result_dict = {
        'deleted_images': len(candidate_image_uuid_list),
        'deleted_annots': len(candidate_annot_uuid_list),
        'species_update': update_dict,
    }
    return result_dict



# -*- coding: utf-8 -*-
import logging
import ubelt as ub
import utool as ut
import warnings

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def have_gpu(min_memory=8000):
    """ Determine if we are on a machine with a good GPU """
    # FIXME: HACK
    gpus = gpu_info()
    if not gpus:
        return False
    return any(gpu['mem_total'] >= min_memory for gpu in gpus)
    # import platform
    # GPU_MACHINES = {'arisia', 'aretha'}
    # # Maybe we look at nvidia-smi instead?
    # hostname = platform.node()
    # return hostname in GPU_MACHINES


def find_unused_gpu(min_memory=0):
    """
    Finds GPU with the lowest memory usage by parsing output of nvidia-smi

    python -c "from pysseg.util import gpu_util; print(gpu_util.find_unused_gpu())"
    """
    gpus = gpu_info()
    if gpus is None:
        return None
    gpu_avail_mem = {n: gpu['mem_avail'] for n, gpu in gpus.items()}
    usage_order = ub.argsort(gpu_avail_mem)
    gpu_num = usage_order[-1]
    if gpu_avail_mem[gpu_num] < min_memory:
        return None
    else:
        return gpu_num


def gpu_info():
    """
    Parses nvidia-smi
    """
    result = ub.cmd('nvidia-smi')
    if result['ret'] != 0:
        warnings.warn('Could not run nvidia-smi.')
        return None

    lines = result['out'].splitlines()

    gpu_lines = []
    current = None

    for line in lines:
        if current is None:
            # Signals the start of GPU info
            if line.startswith('|====='):
                current = []
        else:
            if len(line.strip()) == 0:
                # End of GPU info
                break
            elif line.startswith('+----'):
                # Move to the next GPU
                gpu_lines.append(current)
                current = []
            else:
                current.append(line)

    def parse_gpu_lines(lines):
        line1 = lines[0]
        line2 = lines[1]
        gpu = {}
        gpu['name'] = ' '.join(line1.split('|')[1].split()[1:-1])
        gpu['num'] = int(' '.join(line1.split('|')[1].split()[0]))

        mempart = line2.split('|')[2].strip()
        part1, part2 = mempart.split('/')
        gpu['mem_used'] = float(part1.strip().replace('MiB', ''))
        gpu['mem_total'] = float(part2.strip().replace('MiB', ''))
        gpu['mem_avail'] = gpu['mem_total'] - gpu['mem_used']
        return gpu

    gpus = {}
    for num, lines in enumerate(gpu_lines):
        gpu = parse_gpu_lines(lines)
        assert (
            num == gpu['num']
        ), 'nums ({}, {}) do not agree. probably a parsing error'.format(num, gpu['num'])
        assert (
            num not in gpus
        ), 'Multiple GPUs labeled as num {}. Probably a parsing error'.format(num)
        gpus[num] = gpu
    return gpus



# -*- coding: utf-8 -*-
"""
Unfinished non-wbia dependent version of interact matches
"""
import utool as ut
import six
import numpy as np
from wbia.plottool import abstract_interaction

BASE_CLASS = abstract_interaction.AbstractInteraction


# TODO: move to plottool and decouple with IBEIS
# TODO: abstract interaction
@six.add_metaclass(ut.ReloadingMetaclass)
class MatchInteraction2(BASE_CLASS):
    """
    TODO: replace functional version with this class

    Plots a chip result and sets up callbacks for interaction.

    SeeAlso:
        wbia.viz.interact.interact_matches.MatchInteraction

    CommandLine:
        python -m wbia.plottool.interact_matches --test-MatchInteraction2 --show


    Example:
        >>> # xdoctest: +REQUIRES(module:wbia)
        >>> from wbia.plottool.interact_matches import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> qreq_ = ibs.new_query_request([1], [2, 3, 4, 5], cfgdict=dict(query_rotation_heuristic=True))
        >>> cm = qreq_.execute()[0]
        >>> qaid = cm.qaid
        >>> daid = cm.get_top_aids()[0]
        >>> rchip1 = ibs.get_annot_chips([qaid], config2_=qreq_.extern_query_config2)[0]
        >>> rchip2 = ibs.get_annot_chips([daid], config2_=qreq_.extern_data_config2)[0]
        >>> kpts1 = ibs.get_annot_kpts([qaid], config2_=qreq_.extern_query_config2)[0]
        >>> kpts2 = ibs.get_annot_kpts([daid], config2_=qreq_.extern_data_config2)[0]
        >>> vecs1 = ibs.get_annot_vecs([qaid], config2_=qreq_.extern_query_config2)[0]
        >>> vecs2 = ibs.get_annot_vecs([daid], config2_=qreq_.extern_data_config2)[0]
        >>> fm = cm.aid2_fm[daid]
        >>> fs = cm.aid2_fs[daid]
        >>> fsv = cm.aid2_fsv[daid]
        >>> H1 = cm.aid2_H[daid]
        >>> self = MatchInteraction2(rchip1, rchip2, kpts1, kpts2, fm, fs, fsv,
        >>>                          vecs1, vecs2, H1)
        >>> self.show_page()
        >>> import wbia.plottool as pt
        >>> pt.show_if_requested()

    """

    def __init__(
        self,
        rchip1,
        rchip2,
        kpts1,
        kpts2,
        fm,
        fs,
        fsv,
        vecs1,
        vecs2,
        H1=None,
        H2=None,
        fnum=None,
        **kwargs
    ):
        import wbia.plottool as pt

        kwargs = kwargs.copy()

        # Drawing Data
        self.rchip1 = rchip1
        self.rchip2 = rchip2
        self.kpts1 = kpts1
        self.kpts2 = kpts2
        self.fm = fm
        self.fs = fs
        self.fk = kwargs.pop('fk', None)
        self.fsv = fsv
        self.vecs1 = vecs1
        self.vecs2 = vecs2
        self.H1 = H1
        self.H2 = H2

        # Drawing settings
        self.warp_homog = False
        self.mode = kwargs.pop('mode', 0)
        self.mx = kwargs.pop('mx', None)
        self.vert = kwargs.pop('vert', None)
        self.same_fig = kwargs.get('same_fig', True)
        self.last_fx = 0
        # self.figtitle = kwargs.get('figtitle', 'Inspect Matches')
        self.xywh2 = None
        self.fnum2 = pt.ensure_fnum(fnum)

        self.title = kwargs.get('title', True)
        self.truth = kwargs.pop('truth', None)
        # self.fnum2 = pt.next_fnum()

        # if BASE_CLASS is not object:
        kwargs['interaction_name'] = 'matches'
        super(MatchInteraction2, self).__init__(**kwargs)

        # self.begin(**kwargs)

    def plot(self, *args, **kwargs):
        self.chipmatch_view(*args, **kwargs)

    def chipmatch_view(self, fnum=None, pnum=(1, 1, 1), verbose=None, **kwargs_):
        """
        just visualizes the matches using some type of lines
        """
        import wbia.plottool as pt
        from wbia.plottool import plot_helpers as ph

        if fnum is None:
            fnum = self.fnum
        if verbose is None:
            verbose = ut.VERBOSE

        if verbose:
            print('-- CHIPMATCH VIEW --')
            print('[ichipmatch_view] self.mode = %r' % (self.mode,))
        mode = kwargs_.get('mode', self.mode)
        draw_ell = mode >= 1
        draw_lines = mode == 2
        if verbose:
            print('[ichipmatch_view] draw_lines = %r' % (draw_lines,))
            print('[ichipmatch_view] draw_ell = %r' % (draw_ell,))
        # pt.figure(fnum=fnum, docla=True, doclf=True)
        # NOTE: i remove the clf here. might cause issues
        pt.figure(fnum=fnum, docla=True, doclf=False)
        # show_matches_kw = self.__dict__.copy()
        show_matches_kw = dict(
            # fnum=fnum, pnum=pnum,
            draw_lines=draw_lines,
            draw_ell=draw_ell,
            colorbar_=True,
            vert=self.vert,
            white_background=False,
        )
        show_matches_kw.update(kwargs_)

        if verbose:
            print('self.warp_homog = %r' % (self.warp_homog,))
        if self.warp_homog:
            show_matches_kw['H1'] = self.H1
            show_matches_kw['H2'] = self.H2
        if verbose:
            print('show_matches_kw = %s' % (ut.repr2(show_matches_kw, truncate=True)))

        # tup = show_matches(fm, fs, **show_matches_kw)
        ax, xywh1, xywh2 = pt.show_chipmatch2(
            self.rchip1,
            self.rchip2,
            self.kpts1,
            self.kpts2,
            fm=self.fm,
            fs=self.fs,
            pnum=pnum,
            **show_matches_kw
        )
        self.xywh2 = xywh2
        ph.set_plotdat(ax, 'viztype', 'matches')

        if self.truth is not None and self.truth:
            truth_color = pt.TRUE_BLUE  # if  else pt.FALSE_RED
            pt.draw_border(ax, color=truth_color, lw=4)

        if self.title is not None:
            pt.set_title(self.title, ax=ax)
        # pt.set_figtitle(figtitle + ' ' + vh.get_vsstr(qaid, aid))

    # Draw clicked selection
    def select_ith_match(self, mx):
        """
        Selects the ith match and visualizes and prints information concerning
        features weights, keypoint details, and sift descriptions
        """
        import wbia.plottool as pt
        from wbia.plottool import viz_featrow
        from wbia.plottool import interact_helpers as ih

        fnum = self.fnum
        same_fig = self.same_fig
        rchip1 = self.rchip1
        rchip2 = self.rchip2
        self.mx = mx
        print('+--- SELECT --- ')
        print('... selecting mx-th=%r feature match' % mx)
        fsv = self.fsv
        fs = self.fs
        print('score stats:')
        print(ut.repr2(ut.get_stats(fsv, axis=0), nl=1))
        print('fsv[mx] = %r' % (fsv[mx],))
        print('fs[mx] = %r' % (fs[mx],))
        # ----------------------
        # Get info for the select_ith_match plot
        self.mode = 1
        # Get the mx-th feature match
        fx1, fx2 = self.fm[mx]

        # Older info
        fscore2 = self.fs[mx]
        fk2 = None if self.fk is None else self.fk[mx]
        kp1, kp2 = self.kpts1[fx1], self.kpts2[fx2]
        vecs1, vecs2 = self.vecs1[fx1], self.vecs2[fx2]
        info1 = '\nquery'
        info2 = '\nk=%r fscore=%r' % (fk2, fscore2)
        # self.last_fx = fx1
        self.last_fx = fx1

        # Extracted keypoints to draw
        extracted_list = [
            (rchip1, kp1, vecs1, fx1, 'aid1', info1),
            (rchip2, kp2, vecs2, fx2, 'aid2', info2),
        ]
        # Normalizng Keypoint
        # if hasattr(cm, 'filt2_meta') and 'lnbnn' in cm.filt2_meta:
        #    qfx2_norm = cm.filt2_meta['lnbnn']
        #    # Normalizing chip and feature
        #    (aid3, fx3, normk) = qfx2_norm[fx1]
        #    rchip3 = ibs.get_annot_chips(aid3)
        #    kp3 = ibs.get_annot_kpts(aid3)[fx3]
        #    sift3 = ibs.get_annot_vecs(aid3)[fx3]
        #    info3 = '\nnorm %s k=%r' % (vh.get_aidstrs(aid3), normk)
        #    extracted_list.append((rchip3, kp3, sift3, fx3, aid3, info3))
        # else:
        #    pass
        #    #print('WARNING: meta doesnt exist')

        # ----------------------
        # Draw the select_ith_match plot
        nRows, nCols = len(extracted_list) + same_fig, 3
        # Draw matching chips and features
        sel_fm = np.array([(fx1, fx2)])
        pnum1 = (nRows, 1, 1) if same_fig else (1, 1, 1)
        vert = self.vert if self.vert is not None else False
        self.chipmatch_view(
            pnum=pnum1,
            ell_alpha=0.4,
            ell_linewidth=1.8,
            colors=pt.BLUE,
            sel_fm=sel_fm,
            vert=vert,
        )
        # Draw selected feature matches
        px = nCols * same_fig  # plot offset
        prevsift = None
        if not same_fig:
            # fnum2 = fnum + len(viz.FNUMS)
            fnum2 = self.fnum2
            fig2 = pt.figure(fnum=fnum2, docla=True, doclf=True)
        else:
            fnum2 = fnum

        for (rchip, kp, sift, fx, aid, info) in extracted_list:
            px = viz_featrow.draw_feat_row(
                rchip,
                fx,
                kp,
                sift,
                fnum2,
                nRows,
                nCols,
                px,
                prevsift=prevsift,
                aid=aid,
                info=info,
            )
            prevsift = sift
        if not same_fig:
            ih.connect_callback(fig2, 'button_press_event', self.on_click)
            # pt.set_figtitle(figtitle + vh.get_vsstr(qaid, aid))

    # Callback
    def on_click_inside(self, event, ax):
        from wbia.plottool import plot_helpers as ph

        (x, y) = (event.xdata, event.ydata)
        viztype = ph.get_plotdat(ax, 'viztype', '')

        if event.button == 3:
            self.show_popup_menu(self.get_popup_options(), event)
            return
        # key = '' if event.key is None else event.key
        # ctrl_down = key.find('control') == 0
        if viztype in ['matches', 'multi_match']:
            if len(self.fm) == 0:
                print('[inter] no feature matches to click')
            else:
                # Normal Click
                # Select nearest feature match to the click
                kpts1_m = self.kpts1[self.fm[:, 0]]
                kpts2_m = self.kpts2[self.fm[:, 1]]
                x2, y2, w2, h2 = self.xywh2
                import vtool as vt

                _mx1, _dist1 = vt.nearest_point(x, y, kpts1_m)
                _mx2, _dist2 = vt.nearest_point(x - x2, y - y2, kpts2_m)
                mx = _mx1 if _dist1 < _dist2 else _mx2
                print('... clicked mx=%r' % mx)
                self.select_ith_match(mx)
        # elif viztype in ['warped', 'unwarped']:
        #    pass
        #    #hs_aid = ax.__dict__.get('_hs_aid', None)
        #    #hs_fx = ax.__dict__.get('_hs_fx', None)
        #    #if hs_aid is not None and viztype == 'unwarped':
        #    #    ishow_chip(ibs, hs_aid, fx=hs_fx, fnum=pt.next_fnum())
        #    #elif hs_aid is not None and viztype == 'warped':
        #    #    viz.show_keypoint_gradient_orientations(ibs, hs_aid,
        #    #    hs_fx, fnum=pt.next_fnum())
        # Click in match axes
        # elif viztype == 'matches' and ctrl_down:
        #    # Ctrl-Click
        #    print('.. control click')
        #    return self.sv_view()
        elif viztype.startswith('colorbar'):
            # Hack to get a specific scoring feature
            sortx = self.fs.argsort()
            idx = np.clip(int(np.round(y * len(sortx))), 0, len(sortx) - 1)
            mx = sortx[idx]
            (fx1, fx2) = self.fm[mx]
            (fx1, fx2) = self.fm[mx]
            print('... selected score at rank idx=%r' % (idx,))
            print('... selected score with fs=%r' % (self.fs[mx],))
            print('... resolved to mx=%r' % mx)
            print('... fx1, fx2 = %r, %r' % (fx1, fx2))
            self.select_ith_match(mx)
        else:
            print('...Unknown viztype: %r' % viztype)
        self.draw()

    def on_click_outside(self, event):
        if event.button != 1:
            return
        print('... out of axis')
        # self.warp_homog = not self.warp_homog
        self.mode = (self.mode + 1) % 3
        # self.chipmatch_view()
        self.show_page()
        self.draw()

    def get_popup_options(self):
        def toggle_attr_item(attr, num_states=2):
            value = getattr(self, attr)
            type_ = type(value)

            def toggle_attr():
                new_value = (value + 1) % (num_states)
                new_value = type_(new_value)
                print('new_value(%s) = %r' % (attr, new_value))
                setattr(self, attr, new_value)
                self.show_page()
                self.draw()

            itemstr = 'Toggle %s=%r' % (attr, value)
            return (itemstr, toggle_attr)

        options = [
            toggle_attr_item('warp_homog'),
            toggle_attr_item('mode', 3),
        ]
        return options


def show_keypoint_gradient_orientations(
    ibs, rchip, kp, vec, fnum=None, pnum=None, config2_=None
):
    # Draw the gradient vectors of a patch overlaying the keypoint
    import wbia.plottoola as pt

    if fnum is None:
        fnum = pt.next_fnum()
    # rchip = ibs.get_annot_chips(aid, config2_=config2_)
    # kp    = ibs.get_annot_kpts(aid, config2_=config2_)[fx]
    # sift  = ibs.get_annot_vecs(aid, config2_=config2_)[fx]
    pt.draw_keypoint_gradient_orientations(
        rchip, kp, sift=vec, mode='vec', fnum=fnum, pnum=pnum
    )
    # pt.set_title('Gradient orientation\n %s, fx=%d' % (get_aidstrs(aid), fx))



# -*- coding: utf-8 -*-
"""
TODO: Rename to wbia/init/commands.py

TODO; remove params module
"""
import logging
import utool as ut
import sys
from wbia import constants as const
from wbia import params
from wbia.other import ibsfuncs
from wbia.init import sysres
from os.path import join

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def vdq(dbdir):
    """view directory and quit"""
    _ibsdb = const.PATH_NAMES._ibsdb
    ut.util_cplat.view_directory(join(dbdir, _ibsdb))
    sys.exit(0)


def vdd(ibs):
    """view data dir"""
    ut.util_cplat.view_directory(ibs.dbdir)


def vwd():
    """ view work dir """
    ut.util_cplat.view_directory(sysres.get_workdir())


# def preload_convert_hsdb(dbdir):
#     """ Convert the database before loading (A bit hacky) """
#     from wbia.dbio import ingest_hsdb
#     ingest_hsdb.convert_hsdb_to_wbia(dbdir, force_delete=params.args.force_delete)


def preload_commands(dbdir, **kwargs):
    """ Preload commands work with command line arguments and global caches """
    # logger.info('[main_cmd] preload_commands')
    params.parse_args()
    if params.args.dump_argv:
        logger.info(ut.repr2(vars(params.args), sorted_=False))
    if params.args.dump_global_cache:
        ut.global_cache_dump()  # debug command, dumps to stdout
    if params.args.set_workdir is not None:
        sysres.set_workdir(params.args.set_workdir)
    if params.args.get_workdir:
        logger.info(' Current work dir = %s' % sysres.get_workdir())
    # if params.args.logdir is not None:
    #     sysres.set_logdir(params.args.logdir)
    if params.args.get_logdir:
        logger.info(' Current local  log dir = %s' % (sysres.get_logdir_local(),))
        logger.info(' Current global log dir = %s' % (sysres.get_logdir_global(),))
    if params.args.view_logdir:
        ut.view_directory(sysres.get_logdir_local())
        ut.view_directory(sysres.get_logdir_global())
    if params.args.view_logdir_local:
        ut.view_directory(sysres.get_logdir_local())
    if params.args.view_logdir_global:
        ut.view_directory(sysres.get_logdir_local())
    if ut.get_argflag('--vwd'):
        vwd()
    if ut.get_argflag('--vdq'):
        logger.info('got arg --vdq')
        vdq(dbdir)
    if kwargs.get('delete_ibsdir', False):
        ibsfuncs.delete_wbia_database(dbdir)
    if params.args.preload_exit:
        logger.info('[main_cmd] preload exit')
        sys.exit(0)


def postload_commands(ibs, back):
    """
    Postload commands deal with a specific wbia database

    wbia --db PZ_MTEST --occur "*All Images" --query 1
    wbia --db PZ_MTEST --occur "*All Images" --query-intra

    """
    params.parse_args()
    if ut.NOT_QUIET:
        logger.info('\n[main_cmd] postload_commands')
    if params.args.view_database_directory:
        logger.info('got arg --vdd')
        vdd(ibs)
    if params.args.set_default_dbdir:
        sysres.set_default_dbdir(ibs.get_dbdir())
    if params.args.update_query_cfg is not None:
        # Set query parameters from command line using the --cfg flag
        cfgdict = ut.parse_cfgstr_list(params.args.update_query_cfg)
        logger.info('Custom cfgdict specified')
        logger.info(ut.repr2(cfgdict))
        ibs.update_query_cfg(**cfgdict)
    if params.args.edit_notes:
        ut.editfile(ibs.get_dbnotes_fpath(ensure=True))
    if params.args.delete_cache:
        ibs.delete_cache()
    if params.args.delete_cache_complete:
        ibs.delete_cache(delete_imagesets=True)
    if params.args.delete_query_cache:
        ibs.delete_qres_cache()
    if params.args.set_all_species is not None:
        ibs._overwrite_all_annot_species_to(params.args.set_all_species)
    if params.args.dump_schema:
        ibs.db.print_schema()

    if ut.get_argflag('--ipynb'):
        back.launch_ipy_notebook()

    select_imgsetid = ut.get_argval(
        ('--select-imgsetid', '--imgsetid', '--occur', '--gsid'), None
    )
    if select_imgsetid is not None:
        logger.info('\n+ --- CMD SELECT IMGSETID=%r ---' % (select_imgsetid,))
        # Whoa: this doesnt work. weird.
        # back.select_imgsetid(select_imgsetid)
        # This might be the root of gui problems
        # back.front._change_imageset(select_imgsetid)
        back.front.select_imageset_tab(select_imgsetid)
        logger.info('L ___ CMD SELECT IMGSETID=%r ___\n' % (select_imgsetid,))
    # Send commands to GUIBack
    if params.args.select_aid is not None:
        if back is not None:
            try:
                ibsfuncs.assert_valid_aids(ibs, (params.args.select_aid,))
            except AssertionError:
                logger.info('Valid RIDs are: %r' % (ibs.get_valid_aids(),))
                raise
            back.select_aid(params.args.select_aid)
    if params.args.select_gid is not None:
        back.select_gid(params.args.select_gid)
    if params.args.select_nid is not None:
        back.select_nid(params.args.select_nid)

    select_name = ut.get_argval('--select-name')
    if select_name is not None:
        import wbia.gui.guiheaders as gh

        back.ibswgt.select_table_indicies_from_text(
            gh.NAMES_TREE, select_name, allow_table_change=True
        )

    if ut.get_argflag(('--intra-occur-query', '--query-intra-occur', '--query-intra')):
        back.special_query_funcs['intra_occurrence'](cfgdict={'use_k_padding': False})

    qaid_list = ut.get_argval(('--query-aid', '--query'), type_=list, default=None)

    if qaid_list is not None:
        # qaid_list = params.args.query_aid
        # fix stride case
        if len(qaid_list) == 1 and isinstance(qaid_list[0], tuple):
            qaid_list = list(qaid_list[0])
        daids_mode = ut.get_argval(
            '--daids-mode', type_=str, default=const.VS_EXEMPLARS_KEY
        )
        back.compute_queries(qaid_list=qaid_list, daids_mode=daids_mode, ranks_top=10)

    if ut.get_argflag('--inc-query'):
        back.incremental_query()

    if ut.get_argflag(('--dbinfo', '--display_dbinfo')):
        back.display_dbinfo()
        pass

    aidcmd = ut.get_argval('--aidcmd', default=None)
    aid = ut.get_argval('--aid', type_=int, default=1)
    if aidcmd:
        # aidcmd = 'Interact image'
        metadata = ibs.get_annot_lazy_dict(aid)
        annot_context_options = metadata['annot_context_options']
        aidcmd_dict = dict(annot_context_options)
        logger.info('aidcmd_dict = %s' % (ut.repr3(aidcmd_dict),))
        command = aidcmd_dict[aidcmd]
        command()
        # import utool
        # utool.embed()
        # back.start_web_server_parallel()

    if ut.get_argflag('--start-web'):
        back.start_web_server_parallel()

    if ut.get_argflag('--name-tab'):
        from wbia.gui.guiheaders import NAMES_TREE

        back.front.set_table_tab(NAMES_TREE)
        view = back.front.views[NAMES_TREE]
        model = view.model()
        view._set_sort(model.col_name_list.index('nAids'), col_sort_reverse=True)

    if ut.get_argflag('--graph'):
        back.make_qt_graph_interface()

    if params.args.postload_exit:
        logger.info('[main_cmd] postload exit')
        sys.exit(0)



# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado
"""
import logging
from flask import request, redirect, url_for, current_app
from wbia.control import controller_inject
from wbia.web import appfuncs as appf
from wbia import constants as const
from wbia.web.routes import THROW_TEST_AOI_TURKING
import utool as ut
import numpy as np
import uuid
import six

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

register_route = controller_inject.get_wbia_flask_route(__name__)


@register_route('/submit/login/', methods=['POST'], __route_authenticate__=False)
def submit_login(name, organization, refer=None, *args, **kwargs):
    # Return HTML
    if refer is None:
        refer = url_for('root')
    else:
        refer = appf.decode_refer_url(refer)

    if name == '_new_':
        first = kwargs['new_name_first']
        last = kwargs['new_name_last']
        name = '%s.%s' % (first, last)
        name = name.replace(' ', '')

    if organization == '_new_':
        organization = kwargs['new_organization']
        organization = organization.replace(' ', '')

    name = name.lower()
    organization = organization.lower()

    username = '%s@%s' % (name, organization)
    controller_inject.authenticate(
        username=username, name=name, organization=organization
    )

    return redirect(refer)


@register_route('/submit/cameratrap/', methods=['POST'])
def submit_cameratrap(**kwargs):
    ibs = current_app.ibs
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    gid = int(request.form['cameratrap-gid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)
    flag = request.form.get('cameratrap-toggle', 'off') == 'on'
    ibs.set_image_cameratrap([gid], [flag])
    logger.info('[web] user_id: %s, gid: %d, flag: %r' % (user_id, gid, flag))

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_cameratrap', imgsetid=imgsetid, previous=gid))


@register_route('/submit/detection/', methods=['POST'])
def submit_detection(**kwargs):
    with ut.Timer('submit'):
        is_staged = kwargs.get('staged', False)
        is_canonical = kwargs.get('canonical', False)

        is_staged = is_staged and appf.ALLOW_STAGED

        ibs = current_app.ibs
        method = request.form.get('detection-submit', '')
        imgsetid = request.args.get('imgsetid', '')
        imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
        gid = int(request.form['detection-gid'])
        only_aid = request.form['detection-only-aid']
        only_aid = None if only_aid == 'None' or only_aid == '' else int(only_aid)
        user = controller_inject.get_user()
        if user is None:
            user = {}
        user_id = user.get('username', None)

        if is_canonical:
            assert only_aid is not None
            assert not is_staged
        if only_aid is not None:
            assert is_canonical

        poor_boxes = method.lower() == 'poor boxes'
        if poor_boxes:
            imgsetid_ = ibs.get_imageset_imgsetids_from_text('POOR BOXES')
            ibs.set_image_imgsetids([gid], [imgsetid_])
            method = 'accept'

        if method.lower() == 'delete':
            # ibs.delete_images(gid)
            # logger.info('[web] (DELETED) user_id: %s, gid: %d' % (user_id, gid, ))
            pass
        elif method.lower() == 'clear':
            aid_list = ibs.get_image_aids(gid)
            ibs.delete_annots(aid_list)
            logger.info('[web] (CLEAERED) user_id: %s, gid: %d' % (user_id, gid))
            redirection = request.referrer
            if 'gid' not in redirection:
                # Prevent multiple clears
                if '?' in redirection:
                    redirection = '%s&gid=%d' % (redirection, gid)
                else:
                    redirection = '%s?gid=%d' % (redirection, gid)
            return redirect(redirection)
        elif method.lower() == 'rotate left':
            ibs.update_image_rotate_left_90([gid])
            logger.info('[web] (ROTATED LEFT) user_id: %s, gid: %d' % (user_id, gid))
            redirection = request.referrer
            if 'gid' not in redirection:
                # Prevent multiple clears
                if '?' in redirection:
                    redirection = '%s&gid=%d' % (redirection, gid)
                else:
                    redirection = '%s?gid=%d' % (redirection, gid)
            return redirect(redirection)
        elif method.lower() == 'rotate right':
            ibs.update_image_rotate_right_90([gid])
            logger.info('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, gid))
            redirection = request.referrer
            if 'gid' not in redirection:
                # Prevent multiple clears
                if '?' in redirection:
                    redirection = '%s&gid=%d' % (redirection, gid)
                else:
                    redirection = '%s?gid=%d' % (redirection, gid)
            return redirect(redirection)
        else:
            with ut.Timer('submit...update'):
                current_aid_list = ibs.get_image_aids(gid, is_staged=is_staged)

                if is_canonical:
                    assert (
                        only_aid in current_aid_list
                    ), 'Specified only_aid is not in this image'
                    current_aid_list = [only_aid]

                current_part_rowid_list = ut.flatten(
                    ibs.get_annot_part_rowids(current_aid_list, is_staged=is_staged)
                )

                if is_canonical:
                    current_part_type_list = ibs.get_part_types(current_part_rowid_list)
                    zipped = zip(current_part_rowid_list, current_part_type_list)

                    current_part_rowid_list_ = []
                    for current_part_rowid, current_part_type in zipped:
                        if current_part_type != appf.CANONICAL_PART_TYPE:
                            continue
                        current_part_rowid_list_.append(current_part_rowid)
                    current_part_rowid_list = current_part_rowid_list_

                    assert (
                        len(current_aid_list) == 1
                    ), 'Must have an canonical annotation to focus on in this mode'
                    assert (
                        len(current_part_rowid_list) <= 1
                    ), 'An annotation cannot have more than one canonical part (for now)'

                if is_staged:
                    staged_uuid = uuid.uuid4()
                    staged_user = controller_inject.get_user()
                    if staged_user is None:
                        staged_user = {}
                    staged_user_id = staged_user.get('username', None)

                    # Filter aids for current user
                    current_annot_user_id_list = ibs.get_annot_staged_user_ids(
                        current_aid_list
                    )
                    current_aid_list = [
                        current_aid
                        for current_aid, current_annot_user_id in zip(
                            current_aid_list, current_annot_user_id_list
                        )
                        if current_annot_user_id == staged_user_id
                    ]

                    # Filter part_rowids for current user
                    current_part_user_id_list = ibs.get_part_staged_user_ids(
                        current_part_rowid_list
                    )
                    current_part_rowid_list = [
                        current_part_rowid
                        for current_part_rowid, current_part_user_id in zip(
                            current_part_rowid_list, current_part_user_id_list
                        )
                        if current_part_user_id == staged_user_id
                    ]
                else:
                    staged_uuid = None
                    staged_user = None
                    staged_user_id = None

                # Make new annotations
                width, height = ibs.get_image_sizes(gid)

                if THROW_TEST_AOI_TURKING:
                    # Separate out annotations vs parts
                    raw_manifest = request.form['ia-detection-manifest'].strip()
                    try:
                        manifest_list = ut.from_json(raw_manifest)
                    except ValueError:
                        manifest_list = []
                    test_truth = len(manifest_list) > 0
                    test_challenge_list = [{'gid': gid, 'manifest_list': manifest_list}]
                    test_response_list = [{'poor_boxes': poor_boxes}]
                    test_result_list = [test_truth == poor_boxes]
                    test_user_id_list = [None]
                    ibs.add_test(
                        test_challenge_list,
                        test_response_list,
                        test_result_list=test_result_list,
                        test_user_identity_list=test_user_id_list,
                    )
                else:
                    test_truth = False

                if not test_truth:
                    data_list = ut.from_json(request.form['ia-detection-data'])

                    annotation_list = []
                    part_list = []
                    mapping_dict = {}
                    for outer_index, data in enumerate(data_list):
                        # Check for invalid NaN boxes, filter them out
                        try:
                            assert data['percent']['left'] is not None
                            assert data['percent']['top'] is not None
                            assert data['percent']['width'] is not None
                            assert data['percent']['height'] is not None
                        except Exception:
                            continue

                        parent_index = data['parent']
                        if parent_index is None:
                            inner_index = len(annotation_list)
                            annotation_list.append(data)
                            mapping_dict[outer_index] = inner_index
                        else:
                            assert parent_index in mapping_dict
                            part_list.append(data)

                    ##################################################################################
                    # Get primatives
                    bbox_list = [
                        (
                            int(np.round(width * (annot['percent']['left'] / 100.0))),
                            int(np.round(height * (annot['percent']['top'] / 100.0))),
                            int(np.round(width * (annot['percent']['width'] / 100.0))),
                            int(np.round(height * (annot['percent']['height'] / 100.0))),
                        )
                        for annot in annotation_list
                    ]
                    theta_list = [
                        float(annot['angles']['theta']) for annot in annotation_list
                    ]

                    # Get metadata
                    viewpoint1_list = [
                        int(annot['metadata'].get('viewpoint1', -1))
                        for annot in annotation_list
                    ]
                    viewpoint2_list = [
                        int(annot['metadata'].get('viewpoint2', -1))
                        for annot in annotation_list
                    ]
                    viewpoint3_list = [
                        int(annot['metadata'].get('viewpoint3', -1))
                        for annot in annotation_list
                    ]
                    zipped = zip(viewpoint1_list, viewpoint2_list, viewpoint3_list)
                    viewpoint_list = [
                        appf.convert_tuple_to_viewpoint(tup) for tup in zipped
                    ]

                    quality_list = [
                        int(annot['metadata'].get('quality', 0))
                        for annot in annotation_list
                    ]
                    # Fix qualities
                    for index, quality in enumerate(quality_list):
                        if quality == 0:
                            quality_list[index] = None
                        elif quality == 1:
                            quality_list[index] = 2
                        elif quality == 2:
                            quality_list[index] = 4
                        else:
                            raise ValueError('quality must be 0, 1 or 2')

                    multiple_list = [
                        annot['metadata'].get('multiple', False)
                        for annot in annotation_list
                    ]
                    interest_list = [annot['highlighted'] for annot in annotation_list]
                    species_list = [
                        annot['metadata'].get('species', const.UNKNOWN)
                        for annot in annotation_list
                    ]

                    # Process annotations
                    survived_aid_list = [
                        None if annot['label'] in [None, 'None'] else int(annot['label'])
                        for annot in annotation_list
                    ]

                    # Delete annotations that didn't survive
                    kill_aid_list = list(set(current_aid_list) - set(survived_aid_list))

                    if is_canonical:
                        assert (
                            len(survived_aid_list) == 1
                        ), 'Cannot add or delete annotations in this mode'
                        assert (
                            len(kill_aid_list) == 0
                        ), 'Cannot kill a canonical annotation in this mode'

                    ibs.delete_annots(kill_aid_list)

                    local_add_gid_list = []
                    local_add_bbox_list = []
                    local_add_theta_list = []
                    local_add_interest_list = []
                    local_add_viewpoint_list = []
                    local_add_quality_list = []
                    local_add_multiple_list = []
                    local_add_species_list = []
                    local_add_staged_uuid_list = []
                    local_add_staged_user_id_list = []

                    local_update_aid_list = []
                    local_update_bbox_list = []
                    local_update_theta_list = []
                    local_update_interest_list = []
                    local_update_viewpoint_list = []
                    local_update_quality_list = []
                    local_update_multiple_list = []
                    local_update_species_list = []
                    local_update_staged_uuid_list = []
                    local_update_staged_user_id_list = []

                    zipped = zip(
                        survived_aid_list,
                        bbox_list,
                        theta_list,
                        interest_list,
                        viewpoint_list,
                        quality_list,
                        multiple_list,
                        species_list,
                    )

                    survived_flag_list = []

                    for values in zipped:
                        (
                            aid,
                            bbox,
                            theta,
                            interest,
                            viewpoint,
                            quality,
                            multiple,
                            species,
                        ) = values
                        flag = aid is None
                        survived_flag_list.append(flag)
                        if flag:
                            local_add_gid_list.append(gid)
                            local_add_bbox_list.append(bbox)
                            local_add_theta_list.append(theta)
                            local_add_interest_list.append(interest)
                            local_add_viewpoint_list.append(viewpoint)
                            local_add_quality_list.append(quality)
                            local_add_multiple_list.append(multiple)
                            local_add_species_list.append(species)
                            local_add_staged_uuid_list.append(staged_uuid)
                            local_add_staged_user_id_list.append(staged_user_id)
                        else:
                            local_update_aid_list.append(aid)
                            local_update_bbox_list.append(bbox)
                            local_update_theta_list.append(theta)
                            local_update_interest_list.append(interest)
                            local_update_viewpoint_list.append(viewpoint)
                            local_update_quality_list.append(quality)
                            local_update_multiple_list.append(multiple)
                            local_update_species_list.append(species)
                            local_update_staged_uuid_list.append(staged_uuid)
                            local_update_staged_user_id_list.append(staged_user_id)

                    if len(local_add_gid_list) > 0:
                        add_aid_list = ibs.add_annots(
                            local_add_gid_list,
                            bbox_list=local_add_bbox_list,
                            theta_list=local_add_theta_list,
                            interest_list=local_add_interest_list,
                            viewpoint_list=local_add_viewpoint_list,
                            quality_list=local_add_quality_list,
                            multiple_list=local_add_multiple_list,
                            species_list=local_add_species_list,
                            staged_uuid_list=local_add_staged_uuid_list,
                            staged_user_id_list=local_add_staged_user_id_list,
                            delete_thumb=False,
                        )
                    else:
                        add_aid_list = []

                    if len(local_update_aid_list) > 0:
                        with ut.Timer('submit...update...updating'):
                            ibs.set_annot_bboxes(
                                local_update_aid_list,
                                local_update_bbox_list,
                                theta_list=local_update_theta_list,
                                interest_list=local_update_interest_list,
                            )

                        with ut.Timer('submit...update...metadata0'):
                            ibs.set_annot_viewpoints(
                                local_update_aid_list, local_update_viewpoint_list
                            )
                        with ut.Timer('submit...update...metadata1'):
                            ibs.set_annot_qualities(
                                local_update_aid_list, local_update_quality_list
                            )
                        with ut.Timer('submit...update...metadata2'):
                            ibs.set_annot_multiple(
                                local_update_aid_list, local_update_multiple_list
                            )
                        with ut.Timer('submit...update...metadata3'):
                            ibs.set_annot_species(
                                local_update_aid_list, local_update_species_list
                            )
                        with ut.Timer('submit...update...metadata4'):
                            ibs.set_annot_staged_uuids(
                                local_update_aid_list, local_update_staged_uuid_list
                            )
                        with ut.Timer('submit...update...metadata5'):
                            ibs.set_annot_staged_user_ids(
                                local_update_aid_list, local_update_staged_user_id_list
                            )

                    # Set the mapping dict to use aids now
                    aid_list = []
                    counter_add = 0
                    counter_update = 0
                    for flag in survived_flag_list:
                        if flag:
                            aid_list.append(add_aid_list[counter_add])
                            counter_add += 1
                        else:
                            aid_list.append(local_update_aid_list[counter_update])
                            counter_update += 1

                    assert counter_add == len(add_aid_list)
                    assert counter_update == len(local_update_aid_list)
                    assert counter_add + counter_update == len(survived_aid_list)

                    mapping_dict = {
                        key: aid_list[index] for key, index in mapping_dict.items()
                    }

                    ##################################################################################
                    # Process parts
                    survived_part_rowid_list = [
                        None if part['label'] is None else int(part['label'])
                        for part in part_list
                    ]

                    # Get primatives
                    aid_list = [mapping_dict[part['parent']] for part in part_list]
                    bbox_list = [
                        (
                            int(np.round(width * (part['percent']['left'] / 100.0))),
                            int(np.round(height * (part['percent']['top'] / 100.0))),
                            int(np.round(width * (part['percent']['width'] / 100.0))),
                            int(np.round(height * (part['percent']['height'] / 100.0))),
                        )
                        for part in part_list
                    ]
                    theta_list = [float(part['angles']['theta']) for part in part_list]

                    # Get metadata
                    viewpoint1_list = [
                        int(part['metadata'].get('viewpoint1', -1)) for part in part_list
                    ]
                    viewpoint2_list = [-1] * len(part_list)
                    viewpoint3_list = [-1] * len(part_list)
                    zipped = zip(viewpoint1_list, viewpoint2_list, viewpoint3_list)
                    viewpoint_list = [
                        appf.convert_tuple_to_viewpoint(tup) for tup in zipped
                    ]

                    quality_list = [
                        int(part['metadata'].get('quality', 0)) for part in part_list
                    ]
                    # Fix qualities
                    for index, quality in enumerate(quality_list):
                        if quality == 0:
                            quality_list[index] = None
                        elif quality == 1:
                            quality_list[index] = 2
                        elif quality == 2:
                            quality_list[index] = 4
                        else:
                            raise ValueError('quality must be 0, 1 or 2')

                    type_list = [
                        part['metadata'].get('type', const.UNKNOWN) for part in part_list
                    ]

                    # Delete annotations that didn't survive
                    kill_part_rowid_list = list(
                        set(current_part_rowid_list) - set(survived_part_rowid_list)
                    )

                    if is_canonical:
                        assert (
                            len(survived_part_rowid_list) <= 1
                        ), 'Cannot add more than one canonical part in this mode'
                        assert (
                            len(kill_part_rowid_list) <= 1
                        ), 'Cannot delete two or more canonical parts in this mode'
                        assert len(survived_part_rowid_list) == len(type_list)
                        type_list = [appf.CANONICAL_PART_TYPE] * len(
                            survived_part_rowid_list
                        )

                    ibs.delete_parts(kill_part_rowid_list)

                    staged_uuid_list = [staged_uuid] * len(survived_part_rowid_list)
                    staged_user_id_list = [staged_user_id] * len(survived_part_rowid_list)

                    part_rowid_list = []
                    zipped = zip(
                        survived_part_rowid_list,
                        aid_list,
                        bbox_list,
                        staged_uuid_list,
                        staged_user_id_list,
                    )
                    for part_rowid, aid, bbox, staged_uuid, staged_user_id in zipped:
                        staged_uuid_list_ = None if staged_uuid is None else [staged_uuid]
                        staged_user_id_list_ = (
                            None if staged_user_id is None else [staged_user_id]
                        )

                        if part_rowid is None:
                            part_rowid_ = ibs.add_parts(
                                [aid],
                                [bbox],
                                staged_uuid_list=staged_uuid_list_,
                                staged_user_id_list=staged_user_id_list_,
                            )
                            part_rowid_ = part_rowid_[0]
                        else:
                            ibs._set_part_aid([part_rowid], [aid])
                            ibs.set_part_bboxes([part_rowid], [bbox])
                            if staged_uuid_list_ is not None:
                                ibs.set_part_staged_uuids([part_rowid], staged_uuid_list_)
                            if staged_user_id_list_ is not None:
                                ibs.set_part_staged_user_ids(
                                    [part_rowid], staged_user_id_list_
                                )

                            part_rowid_ = part_rowid
                        part_rowid_list.append(part_rowid_)

                    # Set part metadata
                    logger.info('part_rowid_list = %r' % (part_rowid_list,))
                    ibs.set_part_thetas(part_rowid_list, theta_list)
                    ibs.set_part_viewpoints(part_rowid_list, viewpoint_list)
                    ibs.set_part_qualities(part_rowid_list, quality_list)
                    ibs.set_part_types(part_rowid_list, type_list)

                    ##################################################################################
                    # Process image

                    if is_staged:
                        # Set image reviewed flag
                        metadata_dict = ibs.get_image_metadata(gid)
                        if 'staged' not in metadata_dict:
                            metadata_dict['staged'] = {
                                'sessions': {'uuids': [], 'user_ids': []}
                            }
                        metadata_dict['staged']['sessions']['uuids'].append(
                            str(staged_uuid)
                        )
                        metadata_dict['staged']['sessions']['user_ids'].append(
                            staged_user_id
                        )
                        ibs.set_image_metadata([gid], [metadata_dict])
                    else:
                        ibs.set_image_reviewed([gid], [1])

                        logger.info(
                            '[web] user_id: %s, gid: %d, annots: %d, parts: %d'
                            % (user_id, gid, len(annotation_list), len(part_list))
                        )

        default_list = [
            'autointerest',
            'interest_bypass',
            'metadata',
            'metadata_viewpoint',
            'metadata_quality',
            'metadata_flags',
            'metadata_flags_aoi',
            'metadata_flags_multiple',
            'metadata_species',
            'metadata_label',
            'metadata_quickhelp',
            'parts',
            'modes_rectangle',
            'modes_diagonal',
            'modes_diagonal2',
            'staged',
            'canonical',
        ]
        config = {
            default: kwargs[default] for default in default_list if default in kwargs
        }

        # Return HTML
        refer = request.args.get('refer', '')
        if len(refer) > 0:
            return redirect(appf.decode_refer_url(refer))
        else:
            signature = 'turk_detection_canonical' if is_canonical else 'turk_detection'
            return redirect(
                url_for(
                    signature,
                    imgsetid=imgsetid,
                    previous=gid,
                    previous_only_aid=only_aid,
                    **config,
                )
            )


@register_route('/submit/viewpoint/', methods=['POST'])
def submit_viewpoint(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('viewpoint-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['viewpoint-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    if method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        logger.info('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    if method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        logger.info('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    if method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        logger.info('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        viewpoint = int(request.form['viewpoint-value'])
        viewpoint_text = appf.VIEWPOINT_MAPPING.get(viewpoint, None)
        species_text = request.form['viewpoint-species']
        ibs.set_annot_viewpoints([aid], [viewpoint_text])
        # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint_text])
        ibs.set_annot_species([aid], [species_text])
        logger.info(
            '[web] user_id: %s, aid: %d, viewpoint_text: %s'
            % (user_id, aid, viewpoint_text)
        )
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_viewpoint',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
            )
        )


@register_route('/submit/viewpoint2/', methods=['POST'])
def submit_viewpoint2(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('viewpoint-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['viewpoint-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    if method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        logger.info('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    if method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        logger.info('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    if method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        logger.info('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        if method.lower() == 'ignore':
            viewpoint = 'unknown'
        else:
            # Get metadata
            viewpoint1 = int(kwargs.get('ia-viewpoint-value-1', None))
            viewpoint2 = int(kwargs.get('ia-viewpoint-value-2', None))
            viewpoint3 = int(kwargs.get('ia-viewpoint-value-3', None))
            viewpoint_tup = (
                viewpoint1,
                viewpoint2,
                viewpoint3,
            )
            viewpoint = appf.convert_tuple_to_viewpoint(viewpoint_tup)
        ibs.set_annot_viewpoints([aid], [viewpoint])
        species_text = request.form['viewpoint-species']
        # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint_text])
        ibs.set_annot_species([aid], [species_text])
        logger.info(
            '[web] user_id: %s, aid: %d, viewpoint_text: %s' % (user_id, aid, viewpoint)
        )
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_viewpoint2',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
            )
        )


@register_route('/submit/viewpoint3/', methods=['POST'])
def submit_viewpoint3(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('viewpoint-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['viewpoint-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)

    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    if method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        logger.info('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    if method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        logger.info('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    if method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        logger.info('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)

        if method.lower() == 'ignore':
            ibs.set_annot_viewpoints(
                [aid], ['ignore'], only_allow_known=False, _code_update=False
            )
            viewpoint = 'ignore'
        else:
            # Get metadata
            viewpoint_str = kwargs.get('viewpoint-text-code', '')
            if not isinstance(viewpoint_str, six.string_types) or len(viewpoint_str) == 0:
                viewpoint_str = None

            if viewpoint_str is None:
                viewpoint = const.VIEW.UNKNOWN
            else:
                viewpoint = getattr(const.VIEW, viewpoint_str, const.VIEW.UNKNOWN)
            ibs.set_annot_viewpoint_int([aid], [viewpoint])

        species_text = request.form['viewpoint-species']
        ibs.set_annot_species([aid], [species_text])
        ibs.set_annot_reviewed([aid], [1])
        logger.info(
            '[web] user_id: %s, aid: %d, viewpoint: %s' % (user_id, aid, viewpoint)
        )

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        retval = redirect(appf.decode_refer_url(refer))
    else:
        retval = redirect(
            url_for(
                'turk_viewpoint3',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
            )
        )

    return retval


@register_route('/submit/annotation/', methods=['POST'])
def submit_annotation(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('ia-annotation-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['ia-annotation-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    elif method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        logger.info('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    elif method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        logger.info('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    elif method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        logger.info('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        try:
            viewpoint = int(request.form['ia-annotation-viewpoint-value'])
        except ValueError:
            viewpoint = int(float(request.form['ia-annotation-viewpoint-value']))
        viewpoint_text = appf.VIEWPOINT_MAPPING.get(viewpoint, None)
        species_text = request.form['ia-annotation-species']
        try:
            quality = int(request.form['ia-quality-value'])
        except ValueError:
            quality = int(float(request.form['ia-quality-value']))
        if quality in [-1, None]:
            quality = None
        elif quality == 0:
            quality = 2
        elif quality == 1:
            quality = 4
        else:
            raise ValueError('quality must be -1, 0 or 1')
        ibs.set_annot_viewpoints([aid], [viewpoint_text])
        # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint_text])
        ibs.set_annot_species([aid], [species_text])
        ibs.set_annot_qualities([aid], [quality])
        multiple = 1 if 'ia-multiple-value' in request.form else 0
        ibs.set_annot_multiple([aid], [multiple])
        ibs.set_annot_reviewed([aid], [1])
        logger.info(
            '[web] user_id: %s, aid: %d, viewpoint: %r, quality: %r, multiple: %r'
            % (user_id, aid, viewpoint_text, quality, multiple)
        )
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_annotation',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
            )
        )


@register_route('/submit/annotation/canonical/', methods=['POST'])
def submit_annotation_canonical(samples=200, species=None, version=1, **kwargs):
    ibs = current_app.ibs

    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    assert version in [1, 2, 3]

    aid_list = kwargs['annotation-canonical-aids']
    canonical_list = kwargs['annotation-canonical-highlighted']
    assert len(aid_list) == len(canonical_list)

    # metadata_list = ibs.get_annot_metadata(aid_list)
    # metadata_list_ = []
    # for metadata, highlight in zip(metadata_list, highlight_list):
    #     if 'turk' not in metadata:
    #         metadata['turk'] = {}

    #     if version == 1:
    #         value = highlight
    #     elif version == 2:
    #         value = not highlight
    #     elif version == 3:
    #         value = highlight

    #     metadata['turk']['canonical'] = value
    #     metadata_list_.append(metadata)

    # ibs.set_annot_metadata(aid_list, metadata_list_)

    value_list = []
    for canonical in canonical_list:
        if version == 1:
            value = canonical
        elif version == 2:
            value = not canonical
        elif version == 3:
            value = canonical
        value_list.append(value)

    ibs.set_annot_canonical(aid_list, value_list)

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_annotation_canonical',
                imgsetid=imgsetid,
                samples=samples,
                species=species,
                version=version,
            )
        )


@register_route('/submit/splits/', methods=['POST'])
def submit_splits(**kwargs):
    # ibs = current_app.ibs

    aid_list = kwargs['annotation-splits-aids']
    highlight_list = kwargs['annotation-splits-highlighted']
    assert len(aid_list) == len(highlight_list)

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_splits', aid=None))


@register_route('/submit/species/', methods=['POST'])
def submit_species(**kwargs):
    ibs = current_app.ibs

    method = request.form.get('ia-species-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    previous_species_rowids = request.form.get('ia-species-rowids', None)
    logger.info('Using previous_species_rowids = %r' % (previous_species_rowids,))

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['ia-species-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)

    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    elif method.lower() == 'skip':
        species_text = const.UNKNOWN
        ibs.set_annot_species([aid], [species_text])
        ibs.set_annot_reviewed([aid], [1])
        logger.info('[web] (SKIP) user_id: %s' % (user_id,))
        return redirect(
            url_for(
                'turk_species',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
                previous_species_rowids=previous_species_rowids,
            )
        )
    elif method.lower() in 'refresh':
        logger.info('[web] (REFRESH) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        if '?' in redirection:
            redirection = '%s&refresh=true' % (redirection,)
        else:
            redirection = '%s?refresh=true' % (redirection,)
        return redirect(redirection)
    elif method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        logger.info('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    elif method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        logger.info('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid)
            else:
                redirection = '%s?aid=%d' % (redirection, aid)
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        # species_text = request.form['ia-species-species']
        species_text = kwargs.get('ia-species-value', '')
        if len(species_text) == 0:
            species_text = const.UNKNOWN
        ibs.set_annot_species([aid], [species_text])
        ibs.set_annot_reviewed([aid], [1])

        metadata_dict = ibs.get_annot_metadata(aid)
        if 'turk' not in metadata_dict:
            metadata_dict['turk'] = {}
        metadata_dict['turk']['species'] = user_id
        ibs.set_annot_metadata([aid], [metadata_dict])

        logger.info(
            '[web] user_id: %s, aid: %d, species: %r (%r)'
            % (user_id, aid, species_text, metadata_dict)
        )
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_species',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
                previous_species_rowids=previous_species_rowids,
            )
        )


@register_route('/submit/part/type/', methods=['POST'])
def submit_part_types(**kwargs):
    ibs = current_app.ibs

    method = request.form.get('ia-part-type-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    # IF DECISION NOT IN previous_part_types, REFRESH = TRUE
    previous_part_types = request.form.get('ia-part-types', None)
    logger.info('Using previous_part_types = %r' % (previous_part_types,))

    part_rowid = int(request.form['ia-part-type-part-rowid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)

    refresh = False
    if method.lower() in 'refresh':
        logger.info('[web] (REFRESH) user_id: %s, part_rowid: %d' % (user_id, part_rowid))
        redirection = request.referrer
        if 'part_rowid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&part_rowid=%d' % (redirection, part_rowid)
            else:
                redirection = '%s?part_rowid=%d' % (redirection, part_rowid)
        if '?' in redirection:
            redirection = '%s&refresh=true' % (redirection,)
        else:
            redirection = '%s?refresh=true' % (redirection,)
        return redirect(redirection)
    elif method.lower() == 'rotate left':
        ibs.update_part_rotate_left_90([part_rowid])
        logger.info(
            '[web] (ROTATED LEFT) user_id: %s, part_rowid: %d' % (user_id, part_rowid)
        )
        redirection = request.referrer
        if 'part_rowid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&part_rowid=%d' % (redirection, part_rowid)
            else:
                redirection = '%s?part_rowid=%d' % (redirection, part_rowid)
        return redirect(redirection)
    elif method.lower() == 'rotate right':
        ibs.update_part_rotate_right_90([part_rowid])
        logger.info(
            '[web] (ROTATED RIGHT) user_id: %s, part_rowid: %d' % (user_id, part_rowid)
        )
        redirection = request.referrer
        if 'part_rowid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&part_rowid=%d' % (redirection, part_rowid)
            else:
                redirection = '%s?part_rowid=%d' % (redirection, part_rowid)
        return redirect(redirection)
    else:
        part_type_text = kwargs.get('ia-part-type-value', '')
        ibs.set_part_types([part_rowid], [part_type_text])
        ibs.set_part_reviewed([part_rowid], [1])

        refresh = part_type_text not in previous_part_types
        logger.info(
            '[web] user_id: %s, part_rowid: %d, type: %r'
            % (user_id, part_rowid, part_type_text)
        )
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_part_types',
                imgsetid=imgsetid,
                previous=part_rowid,
                previous_part_types=previous_part_types,
                refresh=refresh,
            )
        )


@register_route('/submit/quality/', methods=['POST'])
def submit_quality(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('quality-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    aid = int(request.form['quality-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        quality = int(request.form['quality-value'])
        ibs.set_annot_qualities([aid], [quality])
        logger.info('[web] user_id: %s, aid: %d, quality: %d' % (user_id, aid, quality))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for(
                'turk_quality',
                imgsetid=imgsetid,
                src_ag=src_ag,
                dst_ag=dst_ag,
                previous=aid,
            )
        )


@register_route('/submit/demographics/', methods=['POST'])
def submit_demographics(species='zebra_grevys', **kwargs):
    ibs = current_app.ibs

    GGR_UPDATE_AGE_FOR_ALL_NAMED_ANOTATIONS = False
    if ibs.dbname in ['GGR-IBEIS', 'GGR2-IBEIS']:
        GGR_UPDATE_AGE_FOR_ALL_NAMED_ANOTATIONS = True

    method = request.form.get('demographics-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    aid = int(request.form['demographics-aid'])
    user = controller_inject.get_user()
    if user is None:
        user = {}
    user_id = user.get('username', None)

    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        logger.info('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid))
        aid = None  # Reset AID to prevent previous
    else:
        sex = int(request.form['demographics-sex-value'])
        age = int(request.form['demographics-age-value'])
        age_min = None
        age_max = None
        # Sex
        if sex >= 2:
            sex -= 2
        else:
            sex = -1

        if age == 1:
            age_min = None
            age_max = None
        elif age == 2:
            age_min = None
            age_max = 2
        elif age == 3:
            age_min = 3
            age_max = 5
        elif age == 4:
            age_min = 6
            age_max = 11
        elif age == 5:
            age_min = 12
            age_max = 23
        elif age == 6:
            age_min = 24
            age_max = 35
        elif age == 7:
            age_min = 36
            age_max = None

        ibs.set_annot_sex([aid], [sex])
        nid = ibs.get_annot_name_rowids(aid)
        if nid is not None and GGR_UPDATE_AGE_FOR_ALL_NAMED_ANOTATIONS:
            aid_list = ibs.get_name_aids(nid)
        else:
            aid_list = [aid]

        ibs.set_annot_age_months_est_min(aid_list, [age_min] * len(aid_list))
        ibs.set_annot_age_months_est_max(aid_list, [age_max] * len(aid_list))
        logger.info(
            '[web] Updating %d demographics with user_id: %s\n\taid_list : %r\n\tsex: %r\n\tage_min: %r\n\tage_max: %r'
            % (len(aid_list), user_id, aid_list, sex, age_min, age_max)
        )
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for('turk_demographics', imgsetid=imgsetid, previous=aid, species=species)
        )


@register_route('/submit/identification/', methods=['POST'])
def submit_identification(**kwargs):
    from wbia.web.apis_query import process_graph_match_html

    ibs = current_app.ibs

    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    aid1 = int(request.form['identification-aid1'])
    aid2 = int(request.form['identification-aid2'])
    replace_review_rowid = int(
        request.form.get('identification-replace-review-rowid', -1)
    )

    # Process form data
    annot_uuid_1, annot_uuid_2, state, tag_list = process_graph_match_html(ibs)

    # Add state to staging database
    # FIXME:
    # photobomb and scenerymatch tags should be disjoint from match-state
    if state == 'matched':
        decision = const.EVIDENCE_DECISION.POSITIVE
    elif state == 'notmatched':
        decision = const.EVIDENCE_DECISION.NEGATIVE
    elif state == 'notcomparable':
        decision = const.EVIDENCE_DECISION.INCOMPARABLE
    elif state == 'photobomb':
        decision = const.EVIDENCE_DECISION.NEGATIVE
        tag_list = ['photobomb']
    elif state == 'scenerymatch':
        decision = const.EVIDENCE_DECISION.NEGATIVE
        tag_list = ['scenerymatch']
    else:
        raise ValueError()

    # Replace a previous decision
    if replace_review_rowid > 0:
        logger.info('REPLACING OLD EVIDENCE_DECISION ID = %r' % (replace_review_rowid,))
        ibs.delete_review([replace_review_rowid])

    # Add a new review row for the new decision (possibly replacing the old one)
    logger.info(
        'ADDING EVIDENCE_DECISION: %r %r %r %r' % (aid1, aid2, decision, tag_list)
    )
    tags_list = None if tag_list is None else [tag_list]
    review_rowid = ibs.add_review([aid1], [aid2], [decision], tags_list=tags_list)
    review_rowid = review_rowid[0]
    previous = '%s;%s;%s' % (aid1, aid2, review_rowid)

    # Notify any attached web QUERY_OBJECT
    try:
        state = const.EVIDENCE_DECISION.INT_TO_CODE[decision]
        feedback = (aid1, aid2, state, tags_list)
        logger.info('Adding %r to QUERY_OBJECT_FEEDBACK_BUFFER' % (feedback,))
        current_app.QUERY_OBJECT_FEEDBACK_BUFFER.append(feedback)
    except ValueError:
        pass

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for('turk_identification', imgsetid=imgsetid, previous=previous)
        )


@register_route('/submit/identification/v2/', methods=['POST'])
def submit_identification_v2(graph_uuid, **kwargs):
    ibs = current_app.ibs

    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    # Process form data
    annot_uuid_1, annot_uuid_2 = ibs.process_graph_match_html_v2(graph_uuid, **kwargs)
    aid1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)

    hogwild = kwargs.get('identification-hogwild', False)
    hogwild_species = kwargs.get('identification-hogwild-species', None)
    hogwild_species = (
        None if hogwild_species == 'None' or hogwild_species == '' else hogwild_species
    )
    logger.info('Using hogwild: %r' % (hogwild,))

    previous = '%s;%s;-1' % (aid1, aid2)

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        base = url_for('turk_identification_graph')
        sep = '&' if '?' in base else '?'
        args = (
            base,
            sep,
            ut.to_json(graph_uuid),
            previous,
            hogwild,
            hogwild_species,
        )
        url = '%s%sgraph_uuid=%s&previous=%s&hogwild=%s&hogwild_species=%s' % args
        url = url.replace(': ', ':')
        return redirect(url)


@register_route('/submit/identification/v2/kaia/', methods=['POST'])
def submit_identification_v2_kaia(graph_uuid, **kwargs):
    ibs = current_app.ibs

    # Process form data
    annot_uuid_1, annot_uuid_2 = ibs.process_graph_match_html_v2(graph_uuid, **kwargs)
    aid1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)

    age1 = kwargs.get('age-annot-1', None)
    age2 = kwargs.get('age-annot-2', None)
    # sex1 = kwargs.get('sex-annot-1', None)
    # sex2 = kwargs.get('sex-annot-2', None)
    condition1 = kwargs.get('condition-annot-1', None)
    condition2 = kwargs.get('condition-annot-2', None)
    comment1 = kwargs.get('comment-annot-1', None)
    comment2 = kwargs.get('comment-annot-2', None)
    comment_match = kwargs.get('comment-match', None)

    assert age1 in ['age1', 'age2', 'age3', 'age4', 'age5', 'age6', 'unknown']
    assert age2 in ['age1', 'age2', 'age3', 'age4', 'age5', 'age6', 'unknown']
    # assert sex1 in ['male', 'female', 'unknown']
    # assert sex2 in ['male', 'female', 'unknown']
    assert 0 <= condition1 and condition1 <= 5
    assert 0 <= condition2 and condition2 <= 5

    # if sex1 == 'male':
    #     sex1 = 1
    # elif sex1 == 'female':
    #     sex1 = 0
    # else:
    #     sex1 = -1

    # if sex2 == 'male':
    #     sex2 = 1
    # elif sex2 == 'female':
    #     sex2 = 0
    # else:
    #     sex2 = -1

    age1_min = None
    age1_max = None
    if age1 == 'age1':
        age1_min = None
        age1_max = 2
    elif age1 == 'age2':
        age1_min = 3
        age1_max = 5
    elif age1 == 'age3':
        age1_min = 6
        age1_max = 11
    elif age1 == 'age4':
        age1_min = 12
        age1_max = 23
    elif age1 == 'age5':
        age1_min = 24
        age1_max = 35
    elif age1 == 'age6':
        age1_min = 36
        age1_max = None
    elif age1 == 'unknown':
        age1_min = None
        age1_max = None
    else:
        raise ValueError()

    age2_min = None
    age2_max = None
    if age2 == 'age1':
        age2_min = None
        age2_max = 2
    elif age2 == 'age2':
        age2_min = 3
        age2_max = 5
    elif age2 == 'age3':
        age2_min = 6
        age2_max = 11
    elif age2 == 'age4':
        age2_min = 12
        age2_max = 23
    elif age2 == 'age5':
        age2_min = 24
        age2_max = 35
    elif age2 == 'age6':
        age2_min = 36
        age2_max = None
    elif age2 == 'unknown':
        age2_min = None
        age2_max = None
    else:
        raise ValueError()

    if condition1 in [0]:
        condition1 = None
    if condition2 in [0]:
        condition2 = None

    # ibs.set_annot_sex([aid1], [sex1])
    # ibs.set_annot_sex([aid2], [sex2])
    ibs.set_annot_age_months_est_min([aid1, aid2], [age1_min, age2_min])
    ibs.set_annot_age_months_est_max([aid1, aid2], [age1_max, age2_max])
    ibs.set_annot_qualities([aid1, aid2], [condition1, condition2])

    metadata1, metadata2 = ibs.get_annot_metadata([aid1, aid2])
    if 'turk' not in metadata1:
        metadata1['turk'] = {}
    if 'turk' not in metadata2:
        metadata2['turk'] = {}
    if 'match' not in metadata1['turk']:
        metadata1['turk']['match'] = {}
    if 'match' not in metadata2['turk']:
        metadata2['turk']['match'] = {}
    metadata1['turk']['match']['comment'] = comment1
    metadata2['turk']['match']['comment'] = comment2
    ibs.set_annot_metadata([aid1, aid2], [metadata1, metadata2])

    edge = (
        aid1,
        aid2,
    )
    review_rowid_list = ibs.get_review_rowids_from_edges([edge])[0]
    if len(review_rowid_list) > 0:
        review_rowid = review_rowid_list[-1]
        metadata_match = ibs.get_review_metadata(review_rowid)
        if 'turk' not in metadata_match:
            metadata_match['turk'] = {}
        if 'match' not in metadata_match['turk']:
            metadata_match['turk']['match'] = {}
        existing_comment = metadata_match.get('comment', '')
        updated_comment = '\n'.join([comment_match, existing_comment])
        updated_comment = updated_comment.strip()
        metadata_match['turk']['match']['comment'] = updated_comment
        ibs.set_review_metadata([review_rowid], [metadata_match])

    hogwild = kwargs.get('identification-hogwild', False)
    hogwild_species = kwargs.get('identification-hogwild-species', None)
    hogwild_species = (
        None if hogwild_species == 'None' or hogwild_species == '' else hogwild_species
    )
    logger.info('Using hogwild: %r' % (hogwild,))

    previous = '%s;%s;-1' % (aid1, aid2)

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        base = url_for('turk_identification_graph')
        sep = '&' if '?' in base else '?'
        args = (
            base,
            sep,
            ut.to_json(graph_uuid),
            previous,
            hogwild,
            hogwild_species,
        )
        url = (
            '%s%sgraph_uuid=%s&previous=%s&hogwild=%s&hogwild_species=%s&kaia=true' % args
        )
        url = url.replace(': ', ':')
        return redirect(url)


@register_route('/submit/group_review/', methods=['POST'])
def group_review_submit(**kwargs):
    """
    CommandLine:
        python -m wbia.web.app --exec-group_review_submit

    Example:
        >>> # UNSTABLE_DOCTEST
        >>> from wbia.web.app import *  # NOQA
        >>> import wbia
        >>> import wbia.web
        >>> ibs = wbia.opendb('testdb1')
        >>> aid_list = ibs.get_valid_aids()[::2]
        >>> ibs.start_web_annot_groupreview(aid_list)
    """
    ibs = current_app.ibs
    method = request.form.get('group-review-submit', '')
    if method.lower() == 'populate':
        redirection = request.referrer
        if 'prefill' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&prefill=true' % (redirection,)
            else:
                redirection = '%s?prefill=true' % (redirection,)
        return redirect(redirection)
    aid_list = request.form.get('aid_list', '')
    if len(aid_list) > 0:
        aid_list = aid_list.replace('[', '')
        aid_list = aid_list.replace(']', '')
        aid_list = aid_list.strip().split(',')
        aid_list = [int(aid_.strip()) for aid_ in aid_list]
    else:
        aid_list = []
    src_ag, dst_ag = ibs.prepare_annotgroup_review(aid_list)
    valid_modes = ut.get_list_column(appf.VALID_TURK_MODES, 0)
    mode = request.form.get('group-review-mode', None)
    assert mode in valid_modes
    return redirect(url_for(mode, src_ag=src_ag, dst_ag=dst_ag))


@register_route('/submit/part/contour/', methods=['POST'])
def submit_contour(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('contour-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    part_rowid = int(request.form['contour-part-rowid'])

    if method.lower() == 'accept':
        data_dict = ut.from_json(request.form['ia-contour-data'])

        if data_dict is None:
            data_dict = {}

        contour_dict = ibs.get_part_contour(part_rowid)
        contour_dict['contour'] = data_dict
        ibs.set_part_contour([part_rowid], [contour_dict])

        segment = data_dict.get('segment', [])
        num_contours = 1
        num_points = len(segment)
        ibs.set_part_reviewed([part_rowid], [1])

        logger.info(
            '[web] part_rowid: %d, contours: %d, points: %d'
            % (part_rowid, num_contours, num_points)
        )

    default_list = ['temp']
    config = {default: kwargs[default] for default in default_list if default in kwargs}

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(
            url_for('turk_contour', imgsetid=imgsetid, previous=part_rowid, **config)
        )



# -*- coding: utf-8 -*-


def dummy_bbox(img, shiftxy=(0.0, 0.0), scale=0.25):
    """Default to rectangle that has a quarter-width/height border."""
    (gh, gw) = img.shape[0:2]
    half_w = gw / 2
    half_h = gh / 2
    w = gw * scale
    h = gh * scale
    x = half_w - (w / 2) + shiftxy[0] * gw
    y = half_h - (h / 2) + shiftxy[1] * gh
    bbox = tuple(map(int, map(round, [x, y, w, h])))
    return bbox


def imread_many(imgpaths):
    import cv2

    img_list = [cv2.imread(img_fpath) for img_fpath in imgpaths]
    return img_list



# -*- coding: utf-8 -*-
import logging
import utool as ut
import vtool as vt
from wbia.viz import viz_helpers as vh
from wbia.algo.detect import randomforest
from os.path import splitext
from wbia.plottool import viz_image2
import wbia.plottool as pt

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def show_hough_image(ibs, gid, species=None, fnum=None, **kwargs):
    if fnum is None:
        fnum = pt.next_fnum()
    title = 'Hough Image: ' + vh.get_image_titles(ibs, gid)
    logger.info(title)

    if species is None:
        species = ibs.get_
        # .cfg.detect_cfg.species_text
    src_gpath_list = ibs.get_image_detectpaths([gid])
    dst_gpath_list = [splitext(gpath)[0] for gpath in src_gpath_list]
    hough_gpath_list = [gpath + '_' + species + '_hough.png' for gpath in dst_gpath_list]
    # Detect with hough
    config = {
        'output_gpath_list': hough_gpath_list,
    }
    logger.info('-' * 80)
    logger.info('')
    logger.info('WARNING!!!')
    logger.info('Hough image is not used often and not worth putting into depcache.')
    logger.info('This image is computed as needed and not cached to disk.')
    logger.info('')
    logger.info('-' * 80)
    results_list = list(  # NOQA
        randomforest.detect_gpath_list_with_species(
            ibs, src_gpath_list, species, **config
        )
    )
    # Get path
    hough_gpath = hough_gpath_list[0]
    img = vt.imread(hough_gpath)
    fig, ax = viz_image2.show_image(img, title=title, fnum=fnum, **kwargs)
    return fig, ax


def show_probability_chip(
    ibs, aid, species=None, fnum=None, config2_=None, blend=False, **kwargs
):
    """
    TODO: allow species override in controller

    CommandLine:
        python -m wbia.viz.viz_hough --exec-show_probability_chip --cnn --show
        python -m wbia.viz.viz_hough --exec-show_probability_chip --cnn --show --db PZ_Master1
        python -m wbia.viz.viz_hough --exec-show_probability_chip --cnn --show --db PZ_Master1 --aid 9970

    Example:
        >>> # SCRIPT
        >>> from wbia.viz.viz_hough import *  # NOQA
        >>> import wbia
        >>> from wbia.viz import viz_chip
        >>> ibs, aid_list, kwargs, config2_ = viz_chip.testdata_showchip()
        >>> fnum = 1
        >>> species = None
        >>> aid = aid_list[0]
        >>> fig, ax = show_probability_chip(ibs, aid, species, fnum, blend=True, **kwargs)
        >>> ut.show_if_requested()
    """
    fnum = pt.ensure_fnum(fnum)
    title = 'Probability Chip: ' + ', '.join(vh.get_annot_text(ibs, [aid], True))
    hough_cpath = ibs.get_annot_probchip_fpath(aid, config2_=config2_)
    img = vt.imread(hough_cpath)
    if blend:
        chip = ibs.get_annot_chips(aid, config2_=config2_)
        img = vt.blend_images_multiply(chip, vt.resize_mask(img, chip))
    fig, ax = viz_image2.show_image(img, title=title, fnum=fnum, **kwargs)
    return fig, ax



# -*- coding: utf-8 -*-
"""
Interface to pydarknet yolo object detection.
"""
import logging
import utool as ut
import vtool as vt
from six.moves import zip

(print, rrr, profile) = ut.inject2(__name__, '[yolo]')
logger = logging.getLogger('wbia')

if not ut.get_argflag('--no-pydarknet'):
    try:
        import pydarknet
    except ImportError:
        logger.info(
            'WARNING Failed to import pydarknet. '
            'PyDarknet YOLO detection is unavailable'
        )
        if ut.SUPER_STRICT:
            raise


VERBOSE_DARK = ut.get_argflag('--verbdark') or ut.VERBOSE


# def train_gid_list(ibs, gid_list, trees_path=None, species=None, setup=True,


def detect_gid_list(ibs, gid_list, downsample=False, **kwargs):
    """
    Args:
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        downsample (bool, optional): a flag to indicate if the original image
            sizes should be used; defaults to True

            True:  ibs.get_image_detectpaths() is used
            False: ibs.get_image_paths() is used

    Kwargs (optional): refer to the PyDarknet documentation for configuration settings

    Args:
        ibs (wbia.IBEISController):  image analysis api
        gid_list (list of int): the list of IBEIS image_rowids that need detection
        downsample (bool, optional): a flag to indicate if the original image
                sizes should be used; defaults to True

    Kwargs:
        detector, config_filepath, weights_filepath, verbose

    Yields:
        tuple: (gid, gpath, result_list)

    CommandLine:
        python -m wbia.algo.detect.yolo detect_gid_list --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.detect.yolo import *  # NOQA
        >>> from wbia.core_images import LocalizerConfig
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='WS_ALL')
        >>> gid_list = ibs.images()._rowids[0:1]
        >>> kwargs = config = LocalizerConfig(**{
        >>>     'weights_filepath': '/media/raid/work/WS_ALL/localizer_backup/detect.yolo.2.39000.weights',
        >>>     'config_filepath': '/media/raid/work/WS_ALL/localizer_backup/detect.yolo.2.cfg',
        >>> })
        >>> exec(ut.execstr_dict(config), globals())
        >>> #classes_fpath = '/media/raid/work/WS_ALL/localizer_backup/detect.yolo.2.cfg.classes'
        >>> downsample = False
        >>> (gid, gpath, result_list) = detect_gid_list(ibs, gid_list, downsample, **config)
        >>> result = ('(gid, gpath, result_list) = %s' % (ut.repr2((gid, gpath, result_list)),))
        >>> print(result)
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> ut.show_if_requested()

    Yields:
        results (list of dict)
    """
    # Get new gpaths if downsampling
    if downsample:
        gpath_list = ibs.get_image_detectpaths(gid_list)
        neww_list = [vt.open_image_size(gpath)[0] for gpath in gpath_list]
        oldw_list = [oldw for (oldw, oldh) in ibs.get_image_sizes(gid_list)]
        downsample_list = [oldw / neww for oldw, neww in zip(oldw_list, neww_list)]
        orient_list = [1] * len(gid_list)
    else:
        gpath_list = ibs.get_image_paths(gid_list)
        downsample_list = [None] * len(gpath_list)
        orient_list = ibs.get_image_orientation(gid_list)
    # Run detection
    results_iter = detect(gpath_list, **kwargs)
    # Upscale the results
    _iter = zip(downsample_list, gid_list, orient_list, results_iter)
    for downsample, gid, orient, (gpath, result_list) in _iter:
        # Upscale the results back up to the original image size
        for result in result_list:
            if downsample is not None and downsample != 1.0:
                for key in ['xtl', 'ytl', 'width', 'height']:
                    result[key] = int(result[key] * downsample)
            bbox = (
                result['xtl'],
                result['ytl'],
                result['width'],
                result['height'],
            )
            bbox_list = [bbox]
            bbox = bbox_list[0]
            result['xtl'], result['ytl'], result['width'], result['height'] = bbox
        yield (gid, gpath, result_list)


def detect(
    gpath_list, detector=None, config_filepath=None, weights_filepath=None, **kwargs
):
    """
    Args:
        gpath_list (list of str): the list of image paths that need detection

    Kwargs (optional): refer to the PyDarknet documentation for configuration settings

    Returns:
        iter

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.detect.yolo import *  # NOQA
        >>> from wbia.core_images import LocalizerConfig
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='WS_ALL')
        >>> gid_list = ibs.images()._rowids[0:1]
        >>> gpath_list = ibs.get_image_paths(gid_list)
        >>> dpath = '/media/raid/work/WS_ALL/localizer_backup/'
        >>> weights_filepath = join(dpath, 'detect.yolo.2.39000.weights')
        >>> config_filepath = join(dpath, 'detect.yolo.2.cfg')
        >>> config = LocalizerConfig(
        >>>     weights_filepath=weights_filepath,
        >>>     config_filepath=config_filepath,
        >>> )
        >>> kwargs = config.asdict()
        >>> ut.delete_dict_keys(kwargs, ['weights_filepath', 'config_filepath'])
        >>> ut.delete_dict_keys(kwargs, ['thumbnail_cfg', 'species', 'algo'])
    """
    # Run detection
    if detector is None:
        classes_filepath = kwargs.pop('classes_filepath', None)
        verbose = kwargs.get('verbose', False)
        detector = pydarknet.Darknet_YOLO_Detector(
            config_filepath=config_filepath,
            weights_filepath=weights_filepath,
            classes_filepath=classes_filepath,
            verbose=verbose,
        )
    # dark = detector
    # input_gpath_list = gpath_list
    results_iter = detector.detect(gpath_list, **kwargs)
    results_list = list(results_iter)
    del detector
    return results_list
